import re

from ..domain import InputPluginAbstract, InputPluginResult

# 类思科的情况
# device>show version
# device(config)# sh version
similar_cisco_reg = re.compile(
    r'^(?P<device_name>[\w\-_\.]+)(?:\(.*?\))?[#|>]\s*(?P<cmd>sh(?:o(?:w)?)?\s+.*)$',
    re.IGNORECASE,
)

# 类华为的情况
# <device>display version
# [device] dis version
simialr_huawei_reg = re.compile(
    r'[\<|\[](?P<device_name>[\w\-_\.]+)(?:\-(?:.+?))?[\]|>]\s*(?P<cmd>dis(?:p(?:l(?:a(?:y)?)?)?)?\s+.*)$',
    re.IGNORECASE,
)

prompt_reg = r'\S+[>|\]|\)|#]'


class InputPluginWithConsole(InputPluginAbstract):
    """通过Console或者vty获取命令的输出"""

    def main(self, file_path: str, stream: str) -> InputPluginResult:
        command = ''
        content = []

        prompt = ''  # 用于记录当前的提示符

        result = InputPluginResult()

        for line in stream.splitlines():
            match = re.match(similar_cisco_reg, line) or re.match(
                simialr_huawei_reg, line
            )  # 判断是否为命令的行
            if match:
                if not result.hostname:  # 如果没有设备名称就记录
                    result.hostname = match.group('device_name')

                prompt = re.match(prompt_reg, line).group(0)  # 获取当前的提示符

                if content and command:  # 如果有内容，且有命令，则保存
                    result.add_cmd(command, '\n'.join(content))

                command = match.group('cmd').strip()
                content.clear()
                continue

            else:  # 如果没有匹配到，有可能是只有命令提示符，此时也要保存
                if prompt:  # 如果保存有提示符
                    # 如果行开始是提示符，且有内容，且有命令，则保存
                    if line.startswith(prompt) and content and command:
                        result.add_cmd(command, '\n'.join(content))

                        command = ''  # 清空状态
                        content.clear()
                        continue

                content.append(line)  # 如果没有匹配到，则添加到内容中

        if content and command:  # 最后将没有保存的内容保存
            result.add_cmd(command, '\n'.join(content))

        return result



from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple

try:
    from openpyxl.styles import Alignment, Border, Font, Side
    from openpyxl.utils import column_index_from_string, get_column_letter
    from openpyxl.workbook import Workbook

    if TYPE_CHECKING:
        from openpyxl.worksheet.worksheet import Worksheet

except ImportError:  # pragma: no cover
    CHECK_IMPORT = False

else:
    CHECK_IMPORT = True


msg = "Please install openpyxl first"


class Excel:
    def __init__(self):
        if not CHECK_IMPORT:
            raise ImportError(msg)  # pragma: no cover

        self.wb, self.sheet = self.init_excel()
        self.next_row = 1  # table的起始行号

    def init_excel(self) -> Tuple[Workbook, Worksheet]:
        """初始化 Excel"""
        wb = Workbook()
        sheet = wb.active
        return wb, sheet

    def save(self, file_path: str):
        self.wb.save(file_path)

    def set_all_column_width(self, width: int, max_col: str = 'A'):
        """设置所有列的宽度"""
        for col in range(1, column_index_from_string(max_col) + 1):
            self.sheet.column_dimensions[get_column_letter(col)].width = width

    def write_row(self, row: List[CellContext | str], merge: List[Tuple[str, str]] = None):
        """写入一行数据"""
        self.write_rows([row], merge)

    def write_rows(
        self,
        rows: List[List[CellContext | str]],
        merge: List[Tuple[str, str]] = None,
    ):
        """
        这个方法是按行来进行写入的, 并且合并操作是每行都会执行一次

        Args:
            - rows: 每行数据的每个数据
            - merge: 合并单元格的范围, 例如: [('A', 'B'), ('C', 'D')]

        """
        # 将 str 转化为 CellContext
        for row_list in rows:
            for i, row in enumerate(row_list):
                if isinstance(row, str):
                    row_list[i] = CellContext(row)


        # 获取要跳过的列
        skip_cols = []
        if merge:
            for merge_item in merge:
                start_col = column_index_from_string(merge_item[0]) + 1
                end_col = column_index_from_string(merge_item[1])
                if start_col == end_col:
                    skip_cols.append(start_col)
                else:
                    skip_cols.extend(range(start_col, end_col + 1))

        for i, row in enumerate(rows):
            col_i = 1
            while row:
                # 跳过合并过的单元格
                if col_i in skip_cols:
                    col_i += 1
                    continue

                current_row = self.next_row + i

                current_cell = self.sheet.cell(current_row, col_i)
                cell = row.pop(0)
                current_cell.value = cell.value
                current_cell.alignment = cell.align
                current_cell.border = cell.border
                current_cell.font = cell.font
                col_i += 1

            if merge is not None:  # 对每行进行合并
                for col_start, col_end in merge:
                    col_start = column_index_from_string(col_start)
                    col_end = column_index_from_string(col_end)
                    self.sheet.merge_cells(
                        start_row=self.next_row + i,
                        start_column=col_start,
                        end_row=self.next_row + i,
                        end_column=col_end,
                    )
        self.next_row += len(rows)  # 更新行号


class CellContext:
    """对单元格设置进行集成"""

    class Style:
        def __init__(self):
            if not CHECK_IMPORT:
                raise ImportError(msg)  # pragma: no cover

            self.align: str = Alignment(horizontal='center', vertical='center')
            self.border: str = Border(
                left=Side(border_style='thin', color='000000'),
                right=Side(border_style='thin', color='000000'),
                top=Side(border_style='thin', color='000000'),
                bottom=Side(border_style='thin', color='000000'),
            )
            self.font: str = Font(
                name='等线', size=10, bold=False, italic=False, strike=False, color='000000'
            )

    def __init__(self, value: str):
        if not CHECK_IMPORT:
            raise ImportError(msg)  # pragma: no cover

        self.value = value
        self.style = self.init_style()

    def init_style(self) -> Style:
        """初始化样式"""
        style = self.Style()
        return style

    @property
    def align(self) -> Alignment:
        return self.style.align

    def set_align(self, align: str) -> Style:
        """设置对齐方式"""
        self.style.align = Alignment(horizontal=align)
        return self

    @property
    def border(self) -> Border:
        return self.style.border

    def set_border(self, border: str) -> Style:
        """设置边框"""
        self.style.border = Border(border_style=border)
        return self

    @property
    def font(self) -> Font:
        return self.style.font

    def set_font(
        self,
        name: str = '等线',
        size: int = 10,
        bold: bool = False,
        italic: bool = False,
        strike: bool = False,
        color: str = '000000',
    ) -> Style:
        """设置字体"""
        self.style.font = Font(
            name=name, size=size, bold=bold, italic=italic, strike=strike, color=color
        )
        return self

    def get_style(self) -> Style:
        """获取样式"""
        return self.style



class Error(Exception):
    ...


class PluginError(Error):
    """插件类错误的基类"""


class TemplateError(PluginError):
    """ntc-template 错误"""


class TemplateNotSupperThisPlatform(TemplateError):
    """ntc-template 不支持该平台"""

    def __init__(self, platform: str):
        self.platform = platform

    def __str__(self):
        return f'platform:{self.platform!r} 暂不支持该平台的解析.'


class TemplateNotSupperThisCommand(TemplateError):
    """ntc-template 不支持该命令的解析"""

    def __init__(self, platform: str, command: str):
        self.platform = platform
        self.command = command

    def __str__(self):
        return f'platform:{self.platform!r} command:{self.command!r} 暂无该平台的textfsm模板文件.'


class NotParseAnyResult(TemplateError):
    """没有解析到结果"""

    def __init__(self, platform: str, command: str):
        self.platform = platform
        self.command = command

    def __str__(self):
        return f'platform:{self.platform!r} cmd:{self.command!r} 未能解析出任何内容, 请检查textfsm模板, 联系仓库维护者或者自定义模板文件.'


class PluginNotSpecify(PluginError):
    """未指定插件"""


class InputPluginError(PluginError):
    """输入插件错误的基类"""


class InputFileTypeError(InputPluginError):
    """输入文件不是期望的形式"""


# FIXME 不存在
class NotFoundDevice(InputPluginError):
    """Input Plugin 中未找到设备"""


class AnalysisPluginError(PluginError):
    """解析插件错误基类"""


class AnalysisLevelError(AnalysisPluginError):
    """分析级别错误"""


# FIXME 不存在
class AnalysisTemplateNameError(AnalysisPluginError):
    """模板的名称不存在"""


# FIXME 不存在
class AnalysisVendorNotSupport(AnalysisPluginError):
    """分析模块暂不支持该厂商"""


class NtcTemplateNotDefined(AnalysisPluginError):
    """模板名称不存在"""


class OutputPluginError(PluginError):
    """输出插件错误基类"""


class OutputParamsNotGiven(OutputPluginError):
    """没有向输出插件提供相应的参数"""

    def __init__(self, plugin_name: str, key: str):
        self.plugin_name = plugin_name
        self.key = key

    def __str__(self):
        return f'没有向输出插件 {self.plugin_name!r} 提供 {self.key!r} 参数'



__version__ = '1.5.2'

import sys

from .api import NetInspect
from .base_info import EachVendorDeviceInfo, BaseInfo, AnalysisInfo
from .domain import (
    InputPluginAbstract,
    OutputPluginAbstract,
    ParsePluginAbstract,
    DeviceInfo,
    DeviceList,
    Device,
    Cmd,
    InputPluginResult,
)

__all__ = [
    "NetInspect",
    "EachVendorDeviceInfo",
    "BaseInfo",
    "AnalysisInfo",
    "InputPluginAbstract",
    "OutputPluginAbstract",
    "ParsePluginAbstract",
    "DeviceInfo",
    "DeviceList",
    "Device",
    "Cmd",
    "InputPluginResult",
]

if sys.version_info < (3, 7):
    print(f'NetInspect {__version__} requires Python 3.7+')
    sys.exit(1)


del sys



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Tuple, Optional

from ..domain import OutputPluginAbstract
from ..third_party.excel import CellContext, Excel

if TYPE_CHECKING:
    from ..domain import Device


class OutputPluginWithExcelReport(OutputPluginAbstract):
    """Microsoft Excel 巡检报告"""

    base_info_keys = [
        ('hostname', '设备名称'),
        ('ip', 'IP'),
        ('vendor', '厂商'),
        ('model', '型号'),
        ('version', '软件版本'),
        ('uptime', '启动时长'),
    ]

    status_info_keys = [
        ('cpu_usage', 'CPU利用率'),
        ('memory_usage', '内存利用率'),
        ('analysis.fan', '风扇状态'),
        ('analysis.power', '电源状态'),
    ]

    @dataclass
    class Values:
        max_column: int = 'F'  # 最大列
        column_width: int = 17  # 列宽
        header_title: str = '现场巡检报告'  # 标题
        header_font: str = '等线'  # 标题字体
        header_font_size: int = 28  # 标题字体大小
        body_font: str = '等线'  # 内容字体
        body_font_size: int = 14  # 内容字体大小
        sn_lines: int = 10  # 序列号至少行数

    def __init__(self):
        self._excel: Optional[Excel] = None
        self.next_row = 1  # table的起始行号

        self.values = self.Values()

    @property
    def excel(self) -> Excel:
        if not self._excel:
            self._excel = Excel()
        
        return self._excel
        

    def main(self):
        """主程序"""
        self.set_values()
        self.set_excel()

        for device in self.args.devices:
            self.report(device)
        self.excel.save(self.args.file_path)

    def set_values(self):
        """可以用于重载设置Value值"""
        pass

    def set_excel(self):
        """对表格的设置"""
        self.excel.set_all_column_width(
            self.values.column_width, self.values.max_column
        )

    def report(self, device: Device):
        """单一报告生成流程"""
        self.report_header(device)
        self.report_body(device)
        self.report_serial_number(device)
        self.report_status(device)

    def report_header(self, device: Device):
        """生成报告头"""
        cell = CellContext(self.values.header_title).set_font(
            name=self.values.header_font, size=self.values.header_font_size
        )
        self.excel.write_rows(rows=[[cell]], merge=[('A', self.values.max_column)])

    def report_body(self, device: Device):
        """生成报告体"""
        self.report_body_baseinfo(device)

    def report_title(self, title: str):
        """生成标题头"""
        cell = CellContext(title).set_font(
            name=self.values.body_font, size=self.values.body_font_size
        )
        self.excel.write_rows(rows=[[cell]], merge=[('A', self.values.max_column)])

    def report_body_baseinfo(self, device: Device):
        """生成基本信息"""
        self.report_title('1. 基本信息')
        rows = self._mulit_keys(self.base_info_keys, device, 2)

        self.excel.write_rows(
            rows=rows, merge=[('B', 'C'), ('E', self.values.max_column)]
        )

    def report_serial_number(self, device: Device):
        """生成序列号"""
        self.report_title('2. 序列号')

        sn_list = device.info.sn
        if self.values.sn_lines > len(sn_list):
            for _ in range(self.values.sn_lines - len(sn_list)):
                sn_list.append(('', ''))

        for pid, sn in sn_list:
            self.excel.write_rows(
                rows=[[CellContext(pid), CellContext(sn)]],
                merge=[('A', 'B'), ('C', self.values.max_column)],
            )

    def report_status(self, device: Device):
        """生成状态信息"""
        self.report_title('3. 状态信息')
        rows = self._mulit_keys(self.status_info_keys, device, 2)
        self.excel.write_rows(
            rows=rows, merge=[('B', 'C'), ('E', self.values.max_column)]
        )

    def _mulit_keys(
        self, keys: List[Tuple[str, str]], device: Device, num: int
    ) -> List[List[CellContext]]:
        """
        每行写入多个值
            Args:
                - keys: 写入的值列表
                - device: 设备对象
                - num: 每行写入的数量

            Return:
                - rows: 写入的值列表
        """
        info = device.info
        rows = []
        for idx, key_tup in enumerate(keys):
            title = CellContext(key_tup[1]).set_font(bold=True)

            # 当遇到analysis.fan这种类型的时候，循环获取analysis.fan的值，并转换为字符串
            if '.' in key_tup[0]:
                key_list = key_tup[0].split('.')
                _value = getattr(info, key_list.pop(0))
                for key in key_list:
                    _value = getattr(_value, key)

                # 如果是bool类型，则显示正常还是异常
                if isinstance(_value, bool):
                    _value = '正常' if _value else '异常'

                value = CellContext(str(_value))
            else:
                value = CellContext(getattr(info, key_tup[0]))

            # 每行写入的数量
            if idx % num == 0:
                rows.append([title, value])
            else:
                rows[-1].extend([title, value])

        return rows



from __future__ import annotations

import os
import re
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List

from ntc_templates.parse import __file__ as model_file
from ntc_templates.parse import get_clitable, parse_output

from .. import exception
from ..domain import ParsePluginAbstract
from ..func import reg_extend

if TYPE_CHECKING:
    from ..domain import Cmd

try:
    from ntc_templates import __author__ as ntc_author

    CHECK_NTC_TEMPLATES = ntc_author == 'elinpf'

except ImportError:
    CHECK_NTC_TEMPLATES = False


class ParsePluginWithNtcTemplates(ParsePluginAbstract):
    """
    使用ntc-templates-elinpf解析命令
    仓库地址：
        https://github.com/Elinpf/ntc-templates
    """

    @dataclass
    class TextFsmInfo:
        """textFSM模板信息"""

        dir: str = ''
        index_commands: Dict[str, List[str]] = field(default_factory=lambda: {})

    def __init__(self):

        if not CHECK_NTC_TEMPLATES:
            msg = (
                "net_inspect 使用的是`ntc_templates_elinpf`这个包，与原`ntc_templates`命名空间重复造成冲突，\n"
                "请删除原`ntc_templates`包: `pip uninstall ntc_templates`,\n"
                "然后安装`ntc_templates_elinpf`包: `pip install ntc_templates_elinpf`"
            )
            raise ImportError(msg)

        ntc_templates_dir = os.path.join(os.path.dirname(model_file), 'templates')
        self.textfms_info_dict = {
            'external': self.TextFsmInfo(),
            'ntc_templates': self.TextFsmInfo(
                dir=ntc_templates_dir,
                index_commands=self._get_index_commands(ntc_templates_dir),
            ),
        }

    def set_external_templates(self, template_dir: str):
        """设置外部的textFSM模板目录
        Args:
            textfsm_dir: textFSM模板目录
        """
        if not os.path.isdir(template_dir):
            raise exception.TemplateError(f'外部模板路径:{template_dir!r} 不存在.')

        self.textfms_info_dict['external'].dir = template_dir
        self.textfms_info_dict['external'].index_commands = self._get_index_commands(
            template_dir
        )

    def _get_index_commands(self, textfsm_dir: str) -> Dict[str, List[str]]:
        """将ntc-templates中的index文件提取出来，
        并将命令按照平台分类
        Args:
            textfsm_dir: textFSM的模板目录
        """

        index_file = os.path.join(textfsm_dir, 'index')

        if not os.path.exists(index_file):
            raise exception.TemplateError(f'外部模板文件夹{textfsm_dir!r}中必须包含`index`文件.')

        commands = {}
        with open(index_file, 'r') as f:
            for line in f.readlines():
                if line == '\n':
                    continue
                elif line.startswith('#'):
                    continue
                elif "Template, Hostname, Platform, Command" in line:
                    continue

                e = line.split(',')
                if len(e) != 4:  # 判断是有效行
                    continue

                platform = e[2].strip()
                if platform not in commands:
                    commands[platform] = []

                commands[platform].append(e[-1].strip())

        return commands

    @contextmanager
    def _pre_parse(self, command: str, platform: str):
        """预处理重复内容

        Args:
            command: 命令
            platform: 命令所属平台
        """
        for type, textfsm_info in self.textfms_info_dict.items():
            if platform not in textfsm_info.index_commands:  # 检查是否有匹配的平台
                if type == 'external':  # 如果是外部的textFSM模板，则进入ntc_templates中
                    continue
                else:
                    raise exception.TemplateNotSupperThisPlatform(platform)

            platform_commands_reg = textfsm_info.index_commands[platform]
            match_command = ''
            for cmd_reg in platform_commands_reg:
                if re.match(f'^{reg_extend(cmd_reg)}$', command):
                    match_command = re.sub(r'[\[|\]]', '', cmd_reg)
                    break

            if match_command == '':  # 检查是否有匹配的命令
                if type == 'external':
                    continue
                else:
                    raise exception.TemplateNotSupperThisCommand(platform, command)

            yield (match_command, textfsm_info)
            break

    def main(self, cmd: Cmd, platform: str) -> Dict[str, str]:
        """识别是否为index中的命令，如果是则使用textFSM解析

        Args:
            cmd: Cmd类
            platform: 命令所属平台, e.g. huawei_os

        Returns:
            解析后的结果
        """
        command = cmd.command

        with self._pre_parse(command, platform) as (match_command, textfsm_info):
            try:
                res = parse_output(
                    platform=platform,
                    command=match_command,
                    data=cmd.content,
                    template_dir=textfsm_info.dir,
                )
            except Exception as e:
                raise exception.TemplateError(
                    f'platform: {platform!r} cmd:{command!r} {str(e)}'
                )

            if not res:  # 如果没有解析到结果，则抛出异常提示
                raise exception.NotParseAnyResult(platform, command)
            return res

    def get_clitable(self, command: str, platform: str) -> dict:
        """通过执行一个空内容的textfsm文件，获得一个字典

        Args:
            command: 命令
            platform: 命令所属平台

        Returns:
            {'command': str, 'platform': str, 'clitable': CliTable}
        """
        with self._pre_parse(command, platform) as (match_command, textfsm_info):
            try:
                clitable = get_clitable(
                    platform=platform, command=command, template_dir=textfsm_info.dir
                )
            except Exception as e:
                raise exception.TemplateError(
                    f'platform: {platform!r} cmd:{command!r} {str(e)}'
                )

            res = {'command': match_command, 'platform': platform, 'clitable': clitable}

            return res



from __future__ import annotations

from .plugins import autoload_plugin, PluginRepository
from .domain import InputPluginAbstract, OutputPluginAbstract, ParsePluginAbstract

from .analysis_plugin import AnalysisPluginAbc


def bootstrap() -> PluginRepository:
    """加载启动项"""
    autoload_plugin()

    plugin_abc = (
        InputPluginAbstract,
        OutputPluginAbstract,
        ParsePluginAbstract,
        AnalysisPluginAbc,
    )
    plugins = []
    for plugin_abstract in plugin_abc:
        temp = plugin_abstract.__subclasses__()
        plugins.append({t.__name__: t for t in temp})

    return PluginRepository(*plugins)



from __future__ import annotations

import os
import re
from typing import TYPE_CHECKING, Dict, List, Optional

from ..data import pystr
from ..domain import PluginAbstract
from ..exception import PluginNotSpecify
from ..logger import logger
from . import __file__ as plugin_file

if TYPE_CHECKING:
    from ..domain import (
        AnalysisPluginAbstract,
        InputPluginAbstract,
        OutputPluginAbstract,
        ParsePluginAbstract,
    )


def autoload_plugin():
    plugin_dir = os.path.dirname(plugin_file)

    for file in os.listdir(plugin_dir):
        if file.endswith(".py") and file != "__init__.py":
            module_name = file[:-3]
            try:
                __import__('.'.join([pystr.software, 'plugins', module_name]))
            except ImportError:
                logger.error('Failed to import plugin: {}'.format(module_name))


Input = 'input'
Output = 'output'
Parse = 'parse'
Analysis = 'analysis'


class PluginRepository:
    """插件仓库，存放插件并且提供插件的获取"""

    def __init__(
        self,
        input_plugins: Dict[str, InputPluginAbstract],
        output_plugins: Dict[str, OutputPluginAbstract],
        parse_plugins: Dict[str, ParsePluginAbstract],
        analysis_plugins: Dict[str, AnalysisPluginAbstract],
    ):
        self.input_plugins = input_plugins
        self.output_plugins = output_plugins
        self.parse_plugins = parse_plugins
        self.analysis_plugins = analysis_plugins
        self._easy_plugin_name = self._to_easy_plugin_name()

    def _to_easy_plugin_name(self) -> Dict[str, PluginAbstract]:
        """变成一个简单的dict，方便查找"""
        ret = {}
        for plugins in [
            self.input_plugins,
            self.output_plugins,
            self.parse_plugins,
            self.analysis_plugins,
        ]:
            for plugin_name, plugin in plugins.items():
                ret[self._lower_name(plugin_name)] = plugin

        return ret

    def _lower_name(self, name: str):
        """将插件名称转换为小写"""
        name = name.lower()
        name = re.sub(r'[-|_]', '', name)
        name = name.replace('.py', '')
        return name

    def get_plugin(self, ptype: str, name: str) -> PluginAbstract:
        """获取插件
        :param ptype: 插件类型
        :param name: 插件名称"""
        lower_name = self._lower_name(name)

        # 支持简写
        if not lower_name.startswith(f"{ptype}pluginwith"):
            lower_name = f"{ptype}pluginwith{lower_name}"

        if lower_name not in self._easy_plugin_name:
            raise PluginNotSpecify('not found plugin: {}'.format(name))

        plugin = self._easy_plugin_name[lower_name]
        if ptype == Input:
            self._check_plugin(self.input_plugins, plugin)

        elif ptype == Output:
            self._check_plugin(self.output_plugins, plugin)

        elif ptype == Parse:
            self._check_plugin(self.parse_plugins, plugin)

        elif ptype == Analysis:
            self._check_plugin(self.analysis_plugins, plugin)

        else:
            raise ValueError('type must be input, output or parse')

        return plugin

    def _check_plugin(
        self, plugin_dict: Dict[str, PluginAbstract], plugin: PluginAbstract
    ):
        """检查这个插件是否在plugin_dict中"""
        if plugin not in plugin_dict.values():
            raise PluginNotSpecify(
                'plugin `{}` not in this plugin type list'.format(plugin)
            )

    def get_input_plugin(self, name: str) -> InputPluginAbstract:
        """获取输入插件
        :param name: 插件名称"""
        return self.get_plugin(Input, name)

    def get_output_plugin(self, name: str) -> OutputPluginAbstract:
        """获取输出插件
        :param name: 插件名称"""
        return self.get_plugin(Output, name)

    def get_parse_plugin(self, name: str) -> ParsePluginAbstract:
        """获取解析插件
        :param name: 插件名称"""
        return self.get_plugin(Parse, name)

    def get_analysis_plugin(self, name: str) -> AnalysisPluginAbstract:
        """获取分析插件
        :param name: 插件名称"""
        return self.get_plugin(Analysis, name)

    def get_analysis_plugin_list(self) -> List[AnalysisPluginAbstract]:
        """获取分析插件列表"""
        return list(self.analysis_plugins.values())

    def plugin_list(self, type: str) -> List[str]:
        """获取名称插件列表
        :param type: 插件类型"""
        if type == Input:
            return list(self.input_plugins.keys())
        elif type == Output:
            return list(self.output_plugins.keys())
        elif type == Parse:
            return list(self.parse_plugins.keys())
        elif type == Analysis:
            return list(self.analysis_plugins.keys())
        else:
            raise ValueError('type must be input, output or parse')

    def input_plugin_list(self) -> List[str]:
        """获取输入插件列表"""
        return self.plugin_list(Input)

    def output_plugin_list(self) -> List[str]:
        """获取输出插件列表"""
        return self.plugin_list(Output)

    def parse_plugin_list(self) -> List[str]:
        """获取解析插件列表"""
        return self.plugin_list(Parse)

    def analysis_plugin_list(self) -> List[str]:
        """获取分析插件列表"""
        return self.plugin_list(Analysis)

    def get_plugins(
        self,
        input_plugin_name: Optional[str] = None,
        output_plugin_name: Optional[str] = None,
        parse_plugin_name: Optional[str] = None,
    ) -> List[Optional[PluginAbstract]]:
        """同时获取三个插件
        :param input_plugin_name: 输入插件名称
        :param output_plugin_name: 输出插件名称
        :param parse_plugin_name: 解析插件名称"""
        return [
            self.get_input_plugin(input_plugin_name) if input_plugin_name else None,
            self.get_output_plugin(output_plugin_name) if output_plugin_name else None,
            self.get_parse_plugin(parse_plugin_name) if parse_plugin_name else None,
        ]



import re
import sys

import rich


def print_err(text: str):
    """打印错误信息"""
    rich.print(f'[[red]x[/]] {text}')


def print_good(text: str):
    """打印成功信息"""
    rich.print(f'[[green]✓[/]] {text}')


def print_warn(text: str):
    """打印警告信息"""
    rich.print(f'[[yellow]![/]] {text}')


def reg_extend(reg: str) -> str:
    """扩展正则表达式写法，支持简单的逐字匹配

    Args:
        reg: 扩展的正则表达式

    Returns:
        str: 还原成可识别的正则表达式

    >>> reg_extend(r'sh[[ow]] ver[[sion]]')
    'sh(o(w)?)? ver(s(i(o(n)?)?)?)?'
    """

    def _completion(match):
        word = str(match.group())[2:-2]
        return '(' + ('(').join(word) + ')?' * len(word)

    return re.sub(r'(\[\[.+?\]\])', _completion, reg)


def get_command_from_textfsm(vendor_platform: str, template: str) -> str:
    """从模板文件名中获得命令

    Args:
        vendor_platform: 平台名称
        template: 模板文件名

    Returns:
        str: 命令

    >>> self.get_command('huawei_os', 'huawei_os_display_interface_status.textfsm')
    'display interface status'
    """

    template = template.replace(vendor_platform + '_', '')
    template = template.replace('.textfsm', '')
    return template.replace('_', ' ')


def pascal_case_to_snake_case(camel_case: str) -> str:
    """大驼峰（帕斯卡）转蛇形

    >>> pascal_case_to_snake_case('HuaweiVrpDisplayVersion')
    'huawei_vrp_display_version'
    """
    snake_case = re.sub(r"(?P<key>[A-Z])", r"_\g<key>", camel_case)
    return snake_case.lower().strip('_')


def snake_case_to_pascal_case(snake_case: str) -> str:
    """蛇形转大驼峰（帕斯卡）

    >>> snake_case_to_pascal_case('huawei_vrp_display_version')
    'HuaweiVrpDisplayVersion'
    """
    words = snake_case.split('_')
    return ''.join(word.title() for word in words)


def clamp_number(num: int, min_num: int, max_num: int) -> int:
    """
    将数字限制在指定范围内

    Args:
        num: 数字
        min_num: 最小值
        max_num: 最大值

    Returns:
        int: 限制后的数字
    """
    return max(min(num, max_num), min_num)


def match(string: str, pattern: str) -> bool:
    """匹配字符串，区分大小写"""
    return bool(re.match(pattern, string))


def match_lower(string: str, pattern: str) -> bool:
    """匹配字符串，不区分大小写"""
    return bool(re.match(pattern, string.lower()))


def safe_float2str(num: float) -> str:
    """安全的将浮点数转换为字符串"""
    return '{:.1f}'.format(num)


def safe_str2float(string: str) -> float:
    """安全的将字符串转换为浮点数"""
    try:
        return float(string)
    except ValueError:
        return 0.0


# 创建一个对大小写不敏感的字典类
class CaseInsensitiveDict(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __setitem__(self, key, value):
        super(CaseInsensitiveDict, self).__setitem__(key.lower(), value)

    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(key.lower())

    def __delitem__(self, key):
        return super(CaseInsensitiveDict, self).__delitem__(key.lower())

    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self[key] = value


class StoreFunc:
    """存储函数的类，用于延迟调用函数"""

    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.func(*self.args, **self.kwargs)


class SkipWithBlock(Exception):
    pass


class Singleton(object):
    """单例类继承"""

    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance


class NoneSkip(Singleton):
    """
    简单模仿None类，但是可以用来作为-with-中的跳过
    使用bool()来判断是否为空
    """

    def __init__(self):
        self.none = None

    def __enter__(self):
        if sys.gettrace():
            return self

        # NOTE 存在一定的问题，当程序处于调试状态的时候，调试进程会被破坏：
        # https://pydev.blogspot.com/2007/06/why-cant-pydev-debugger-work-with.html
        sys.settrace(lambda *args, **keys: None)
        frame = sys._getframe(1)
        frame.f_trace = self.trace

    def trace(self, frame, event, arg):
        raise SkipWithBlock()

    def __bool__(self):
        return False

    def __exit__(self, type, value, traceback):
        if type is None:
            return  # No exception
        if issubclass(type, SkipWithBlock):
            return True  # Suppress special SkipWithBlock exception

    def __eq__(self, other):
        return self.none == other

    def __ne__(self, other):
        return self.none != other

    def __getattr__(self, __name: str):
        # NOTE 当处于Debug状态的时候，调用不存在的属性时，就会触发这个，跳过-with-后面的内容
        raise SkipWithBlock()



from __future__ import annotations
from dataclasses import dataclass, field
import re

from typing import TYPE_CHECKING, Callable, List, Tuple, Optional
from .func import match_lower, Singleton
from .logger import logger


if TYPE_CHECKING:
    from .domain import Device
    from . import vendor


@dataclass
class AnalysisInfo:
    cpu: Optional[bool] = None  # CPU利用率是否异常
    memory: Optional[bool] = None  # Memory 利用率是否异常
    fan: Optional[bool] = None  # 风扇是否异常
    power: Optional[bool] = None  # 电源是否异常

    def __getitem__(self, item: str):
        return self.__getattribute__(item)


@dataclass
class BaseInfo:
    hostname: str = ''  # 主机名
    file_path: str = ''  # 文件路径
    vendor: str = ''  # 厂商名称
    vendor_platform: str = ''  # 厂商软件平台
    model: str = ''  # 型号
    version: str = ''  # 版本
    uptime: str = ''  # 启动时间
    ip: str = ''  # IP地址
    sn: List[Tuple[str, str]] = field(default_factory=lambda: [])  # 序列号
    cpu_usage: str = ''  # CPU使用率
    memory_usage: str = ''  # 内存使用率
    analysis: AnalysisInfo = None  # 检查项目结果

    def __getitem__(self, item: str):
        return self.__getattribute__(item)


class EachVendorDeviceInfo(Singleton):
    """
    这个类的作用是实现各个厂商设备的基本信息获取，将获取到的信息转换格式保存到BaseInfo类中

    ``base_info_class`` , ``analysis_info_class`` , ``append_analysis_items`` 均可以重载

    自行写入方法::

        def do_<vendor_platform>_baseinfo_<func_name>(device: Device, base_info: BaseInfo):
            ...

    """

    analysis_items = [
        ('cpu_status', 'cpu'),
        ('memory_status', 'memory'),
        ('fan_status', 'fan'),
        ('power_status', 'power'),
    ]

    # 可用于重载的信息
    base_info_class = BaseInfo
    analysis_info_class = AnalysisInfo
    append_analysis_items = []

    def run_general_information(self, device: Device) -> BaseInfo:
        """配置通用基础信息"""
        base_info = self.base_info_class()
        base_info.hostname = device._device_info.name
        base_info.file_path = device._device_info.file_path
        base_info.vendor = str(device.vendor).split('.')[-1][:-2]
        base_info.vendor_platform = device.vendor.PLATFORM
        base_info.ip = device._device_info.ip
        base_info.sn = []
        base_info.analysis = self.analysis_info_class()

        return base_info

    def run_baseinfo_func(self, device: Device) -> BaseInfo:
        """获取基本信息"""
        base_info = self.run_general_information(device)

        platform = device.vendor
        funcs = self.get_funcs(platform, 'baseinfo')
        if not funcs:
            logger.warning(f'{platform} baseinfo is not implemented')

        for func in funcs:
            func(device, base_info)

        return base_info

    def get_funcs(
        self, platform: vendor.DefaultVendor, type_name: str
    ) -> List[Callable[[Device, BaseInfo]]]:
        """取设备厂商所有的方法"""
        ret = []
        for i in dir(self):
            if re.match(f'do_{platform.PLATFORM}_{type_name}.*', i):
                ret.append(getattr(self, i))
        return ret

    def do_huawei_vrp_baseinfo(self, device: Device, info: BaseInfo):
        """获取华为设备基本信息"""

        # Manager IP
        if not info.ip:
            with device.search_cmd('display ip interface brief') as cmd:
                for row in cmd.parse_result:
                    if match_lower(row.get('interface'), 'loopback0'):
                        info.ip = row.get('ip')
                        break  # 取到后可以立即返回

                    # 有些设备没有loopback0, 用 vlanif 代替
                    elif match_lower(row.get('interface'), 'vlanif'):
                        if not info.ip:
                            info.ip = row.get('ip')

        # Model & Version & Uptime
        with device.search_cmd('display version') as cmd:
            for row in cmd.parse_result:
                info.model = row.get('model')
                info.version = row.get('vrp_version')
                info.uptime = row.get('uptime')

        # 获取序列号
        with device.search_cmd('display device manufacture-info') as cmd:
            for row in cmd.parse_result:
                info.sn.append((row.get('type'), row.get('serial')))

        # 如果没有获取到序列号，就从 display esn 和 display elabel brief里面找
        if not info.sn:
            with device.search_cmd('display esn') as cmd:
                if cmd.parse_result:
                    sn = cmd.parse_result[0].get('esn')
                    info.sn.append(('chassis', sn))

            with device.search_cmd('display elabel brief') as cmd:
                for row in cmd.parse_result:
                    info.sn.append((row.get('slot'), row.get('bar_code')))

        # CPU利用率
        with device.search_cmd('display cpu-usage') as cmd:
            if cmd.parse_result:
                info.cpu_usage = cmd.parse_result[0].get('cpu_5_min') + '%'

        # Memory 利用率
        with device.search_cmd('display memory-usage') as cmd:
            if cmd.parse_result:
                info.memory_usage = (
                    cmd.parse_result[0].get('memory_using_percent') + '%'
                )

        device.analysis_result

    def do_hp_comware_baseinfo(self, device: Device, info: BaseInfo):
        """获取华三设备基本信息"""

        # Manager IP
        if not info.ip:
            with device.search_cmd('display ip interface brief') as cmd:
                for row in cmd.parse_result:
                    if match_lower(row.get('interface'), 'loop0|loopback0'):
                        info.ip = row.get('ip')
                        break

                    # 有些设备没有loopback0, 用 vlanif 代替
                    elif match_lower(row.get('interface'), 'vlan'):
                        if not info.ip:
                            info.ip = row.get('ip')

        # Model & Version & Uptime
        with device.search_cmd('display version') as cmd:
            for row in cmd.parse_result:
                info.model = row.get('model')
                info.version = (
                    row.get('software_version') + ' Release: ' + row.get('release')
                )
                info.uptime = row.get('uptime')

        # SN
        with device.search_cmd('display device manuinfo') as cmd:
            for row in cmd.parse_result:
                if row.get('device_serial_number').lower() == 'none':
                    continue
                info.sn.append(
                    (row.get('device_name'), row.get('device_serial_number'))
                )

        # CPU利用率
        with device.search_cmd('display cpu-usage') as cmd:
            if cmd.parse_result:
                info.cpu_usage = cmd.parse_result[0].get('cpu_5_min') + '%'

        # Memory 利用率
        with device.search_cmd('display memory') as cmd:
            if cmd.parse_result:
                row = cmd.parse_result[0]
                if row.get('used_rate'):
                    info.memory_usage = row.get('used_rate') + '%'

                # 当是free rate 的情况，需要转换
                elif row.get('free_rate'):
                    info.memory_usage = f"{(100 - float(row.get('free_rate'))):.1f}%"

    def do_maipu_mypower_baseinfo(self, device: Device, info: BaseInfo):
        """获取迈普设备基本信息"""

        # Manager IP
        if not info.ip:
            with device.search_cmd('show ip interface brief') as cmd:
                for row in cmd.parse_result:
                    if match_lower(row.get('interface'), 'loopback0'):
                        info.ip = row.get('ip')
                        break

                    elif match_lower(row.get('interface'), 'vlan'):
                        if not info.ip:
                            info.ip = row.get('ip')

        # Model & Version & Uptime
        with device.search_cmd('show version') as cmd:
            for row in cmd.parse_result:
                info.model = row.get('model')
                info.version = row.get('software_version')
                info.uptime = row.get('uptime')

        with device.search_cmd('show system module brief') as cmd:
            for row in cmd.parse_result:
                if row.get('name') != '/':
                    info.sn.append((row.get('name'), row.get('sn')))

        # CPU 利用率
        with device.search_cmd('show cpu monitor') as cmd:
            if cmd.parse_result:
                info.cpu_usage = cmd.parse_result[0].get('cpu_5_min') + '%'

        # Memory 利用率
        with device.search_cmd('show memory') as cmd:
            if cmd.parse_result:
                info.memory_usage = cmd.parse_result[0].get('used_percent') + '%'

    def do_ruijie_os_baseinfo(self, device: Device, info: BaseInfo):
        """获取锐捷设备基本信息"""

        # Manager IP
        if not info.ip:
            with device.search_cmd('show ip interface brief') as cmd:
                for row in cmd.parse_result:
                    if match_lower(row.get('interface'), 'loopback 0'):
                        info.ip = row.get('ip')
                        break

                    elif match_lower(row.get('interface'), 'vlan'):
                        if not info.ip:
                            info.ip = row.get('ip')

        # Model & Version & Uptime
        with device.search_cmd('show version') as cmd:
            if cmd.parse_result:
                temp = cmd.parse_result[0]
                info.model = temp.get('model')
                info.version = temp.get('soft_version')
                info.uptime = temp.get('uptime')

            # SN
            for row in cmd.parse_result:
                if row.get('serial_number'):
                    info.sn.append((row.get('slot_name'), row.get('serial_number')))

        # CPU 利用率
        with device.search_cmd('show cpu') as cmd:
            if cmd.parse_result:
                info.cpu_usage = cmd.parse_result[0].get('cpu_5_min') + '%'

        # Memory 利用率
        with device.search_cmd('show memory') as cmd:
            if cmd.parse_result:
                info.memory_usage = (
                    cmd.parse_result[0].get('system_memory_used_rate_precent') + '%'
                )

    def do_cisco_ios_baseinfo(self, device: Device, info: BaseInfo):
        """获取思科设备基本信息"""

        # Manager IP
        if not info.ip:
            with device.search_cmd('show ip interface brief') as cmd:
                for row in cmd.parse_result:
                    if match_lower(row.get('intf'), 'loopback0'):
                        info.ip = row.get('ipaddr')
                        break

                    elif match_lower(row.get('intf'), 'vlan'):
                        if not info.ip:
                            info.ip = row.get('ipaddr')

        # Model & Version & Uptime
        with device.search_cmd('show version') as cmd:
            for row in cmd.parse_result:
                info.model = row.get('hardware')[0]
                info.version = row.get('version')
                info.uptime = row.get('uptime')

        # SN
        with device.search_cmd('show inventory') as cmd:
            for row in cmd.parse_result:
                info.sn.append((row.get('name'), row.get('sn')))

        # CPU 利用率
        with device.search_cmd('show processes cpu') as cmd:
            if cmd.parse_result:
                info.cpu_usage = cmd.parse_result[0].get('cpu_5_min') + '%'

        # Memory 利用率
        with device.search_cmd('show processes memory') as cmd:
            if cmd.parse_result:
                total = cmd.parse_result[0].get('memory_total')
                used = cmd.parse_result[0].get('memory_used')
                info.memory_usage = (
                    str(round(float(used) / float(total) * 100, 1)) + '%'
                )

        if not info.memory_usage:  # 有些设备没有show processes memory命令
            for row in device.parse_result('show processes memory sorted'):
                total = row.get('memory_total')
                used = row.get('memory_used')
                info.memory_usage = (
                    str(round(float(used) / float(total) * 100, 1)) + '%'
                )

    def run_analysis_info(self, device: Device):
        """
        更新设备检查信息, 重载追加的内容也会添加进来

        如果没有检查信息，则为 ``None`` ，
        如果检查信息为告警级别，则为 ``True``
        如果检查信息为正常或者关注级别，则为 ``False``
        """
        info = device.info

        for item in self.analysis_items + self.append_analysis_items:
            ar = device.analysis_result.get(item[0])
            if not ar._result:  # 如果没有检查结果，则不更新
                continue
            elif ar.include_warning:  # 如果有检查结果，但是包含warning，则更新为True
                setattr(info.analysis, item[1], True)
            else:  # 如果检查的结果不包含warning， 则更新为False
                setattr(info.analysis, item[1], False)


def get_base_info(device: Device, device_info_handler=EachVendorDeviceInfo) -> BaseInfo:
    """获取设备基本信息"""
    info = (
        device_info_handler()
        if type(device_info_handler) == type
        else device_info_handler
    )
    base_info = info.run_baseinfo_func(device)
    info.run_analysis_info(device)
    return base_info



import sys

from loguru import logger as loguru_logger

from .func import Singleton

logger = loguru_logger.opt(colors=True)


class LoggerConfig(Singleton):
    def __init__(self):
        self._logger = loguru_logger
        self._enable_file_log = False
        self._ebable_console_log = False

        self.remove()

    def remove(self):
        self._logger.remove()
        self._enable_file_log = False
        self._ebable_console_log = False

    def enable_file_log(
        self,
        file_path: str,
        level: str,
        rotation: str,
        log_format: str,
    ):
        if self._enable_file_log is False:
            self._logger.add(
                sink=file_path,
                level=level,
                rotation=rotation,
                format=log_format,
                enqueue=True,
                backtrace=True,
                diagnose=True,
            )
            self._enable_file_log = True

    def enable_console_log(
        self,
        level: str,
        log_format: str,
    ):
        if self._ebable_console_log is False:
            self._logger.add(
                sink=sys.stderr,
                level=level,
                format=log_format,
                enqueue=True,
                backtrace=True,
                diagnose=True,
            )
            self._ebable_console_log = True



from __future__ import annotations

import os
from typing import Dict, List, Tuple

from .data import pyoption
from .domain import DeviceInfo, PluginManagerAbc
from .exception import InputFileTypeError
from .logger import logger


class PluginManager(PluginManagerAbc):
    def input_dir(
        self, dir_path: str, expend: str | List = None
    ) -> List[Tuple[Dict[str, str], DeviceInfo]]:
        """对目录中的文件进行设备输入"""
        devices = []

        if expend is None:
            expend = pyoption.input_file_expend
        elif expend is str:
            expend = [expend]

        for root, dirs, files in os.walk(dir_path):
            for file in files:
                if os.path.splitext(file)[-1] in expend:  # 判断文件后缀
                    file_path = os.path.join(root, file)
                    try:
                        devices.append(self.input(file_path))
                    except InputFileTypeError:  # pragma: no cover
                        logger.info('文件不符合input_plugin标准，跳过: {}'.format(file_path))

        return devices



from __future__ import annotations

from typing import TYPE_CHECKING

from .. import vendor
from ..analysis_plugin import AnalysisPluginAbc, analysis

if TYPE_CHECKING:
    from ..analysis_plugin import TemplateInfo
    from ..domain import AnalysisResult


class AnalysisPluginWithLoggerSummary(AnalysisPluginAbc):
    """
    检查日志中的告警信息。
    """

    @analysis.vendor(vendor.H3C)
    @analysis.template_key(
        'hp_comware_display_logbuffer_summary.textfsm',
        ['slot', 'emerg', 'alert', 'crit', 'error', 'warn'],
    )
    def hp_comware(template: TemplateInfo, result: AnalysisResult):
        """日志摘要中大于WARN级别的告警信息进行提示"""
        for row in template['display logbuffer summary']:
            if int(row['emerg']) > 0:
                result.add_warning(f"Slot {row['slot']} 有{row['emerg']}条紧急告警")
            if int(row['alert']) > 0:
                result.add_warning(f"Slot {row['slot']} 有{row['alert']}条警告告警")
            if int(row['crit']) > 0:
                result.add_warning(f"Slot {row['slot']} 有{row['crit']}条严重告警")
            if int(row['error']) > 0:
                result.add_warning(f"Slot {row['slot']} 有{row['error']}条错误告警")
            if int(row['warn']) > 0:
                result.add_warning(f"Slot {row['slot']} 有{row['warn']}条告警")



from __future__ import annotations

from typing import TYPE_CHECKING

from .. import vendor
from ..analysis_plugin import AnalysisPluginAbc, analysis
from ..func import match_lower

if TYPE_CHECKING:
    from ..analysis_plugin import TemplateInfo
    from ..domain import AnalysisResult


class AnalysisPluginWithFanStatus(AnalysisPluginAbc):
    """
    要求设备所有在位风扇模块运行在正常状态。
    """

    @analysis.vendor(vendor.Huawei)
    @analysis.template_key(
        'huawei_vrp_display_fan.textfsm', ['slot_id', 'present', 'registered', 'status']
    )
    def huawei_vrp(template: TemplateInfo, result: AnalysisResult):
        """检查设备所有在位风扇模块运行在正常状态"""
        for row in template['display fan']:
            if row['present']:  # NE serial and S5700 serial
                if not match_lower(row['present'], r'yes|present'):
                    result.add_focus(f"Slot {row['slot_id']} 风扇不在位")

                elif row['registered'] and not match_lower(row['registered'], 'yes'):
                    result.add_focus(f"Slot {row['slot_id']} 风扇未注册")

                elif not match_lower(row['status'], r'auto|namual|normal'):
                    result.add_warning(f"Slot {row['slot_id']} 风扇状态不正常")
            else:  # 其他系列进行简单的判断
                if not match_lower(row['status'], 'normal'):
                    result.add_warning(f"Slot {row['slot_id']} 风扇状态不正常")

    @analysis.vendor(vendor.H3C)
    @analysis.template_key('hp_comware_display_fan.textfsm', ['slot', 'id', 'status'])
    def hp_comware(template: TemplateInfo, result: AnalysisResult):
        """模块状态不为Normal的时候告警"""
        for row in template['display fan']:
            if not match_lower(row['status'], 'normal'):
                result.add_warning(
                    f'Slot {row["slot"]} Fan {row["id"]} 状态异常'
                    if row['slot']
                    else f'Fan {row["id"]} 状态异常'
                )

    @analysis.vendor(vendor.Maipu)
    @analysis.template_key(
        'maipu_mypower_show_system_fan.textfsm',
        ['fan_id', 'status', 'work_status', 'statistics_ierr', 'statistics_oerr'],
    )
    def maipu_mypower(template: TemplateInfo, result: AnalysisResult):
        """模块Status状态为Online的时候，检查WorkStatus不为Normal的时候告警, 否则Status不为Normal时警告，
        并且检查模块的错误统计"""
        for row in template['show system fan']:
            if match_lower(row['status'], 'online'):
                if not match_lower(row['work_status'], 'normal'):
                    result.add_warning(f'Fan {row["fan_id"]} 状态异常')
            elif not match_lower(row['status'], 'normal'):
                result.add_warning(f'Fan {row["fan_id"]} 状态异常')

            elif int(row['statistics_ierr']) + int(row['statistics_oerr']) > 0:
                result.add_warning(f'Fan {row["fan_id"]} 状态异常')



from rich import print
from rich.table import Table
from ..domain import OutputPluginAbstract


class OutputPluginWithDeviceList(OutputPluginAbstract):
    """
    输出设备列表信息
    """

    def main(self):
        table = Table(title='设备列表')
        columns = ['设备名称', '厂商', '型号', 'IP', 'CPU %', 'Mem %']
        for col in columns:
            table.add_column(col, justify='center')
        table.row_styles = ['green']

        for device in self.args.devices:
            info = device.info
            row = [
                info.hostname,
                info.vendor,
                info.model,
                info.ip,
                info.cpu_usage,
                info.memory_usage,
            ]
            table.add_row(*row)

        print(table)
        print('total devices:', len(self.args.devices))



import os
import re
from typing import Dict, Tuple

from ..domain import DeviceInfo, InputPluginAbstract, InputPluginResult
from ..exception import InputFileTypeError

"""
这个插件是分析的从OSmartOne平台获取的输入文件，有两种情况

情况一：
文件名格式为: A_FOO_BAR_DR01_127.0.0.1_20220221180010.diag
文件内容格式为:
-------------------------show version-------------------------

H3C Comware Platform Software
Comware Software, Version 5.20, Release 2202
Copyright (c) 2004-2010 Hangzhou H3C Tech. Co., Ltd. All rights reserved.
H3C S5500-28C-EI uptime is 1 weeks, 1 day, 1 hours, 1 minutes

H3C S5500-28C-EI with 1 Processor
256M    bytes SDRAM
32768K  bytes Flash Memory

情况二：
文件名格式为: A_FOO_BAR_DR01_127.0.0.1.diag
文件内容格式为：
------------------------------------------------------------
dis version
Huawei Versatile Routing Platform Software
VRP (R) software, Version 8.180 (NE20E V800R010C10SPC500)
Copyright (C) 2012-2018 Huawei Technologies Co., Ltd.
HUAWEI NE20E-S4 uptime is 169 days, 13 hours, 50 minutes 
Patch Version: V800R010SPH120

NE20E-S4 version information:
"""

device_info_reg = r'(?P<name>\S+)\_(?P<ip>(?:\d+\.){3}(?:\d+)).*'
command_line_reg = (
    r'^-------------------------(?P<cmd>[^-].*?)-------------------------$'
)
command_line_reg2 = r'^------------------------------------------------------------$'


class InputPluginWithSmartOne(InputPluginAbstract):
    """通过iSmartOne平台获取的输出"""

    def main(self, file_path: str, stream: str) -> InputPluginResult:

        result = InputPluginResult()

        match = re.match(device_info_reg, os.path.basename(file_path))

        if not match:
            raise InputFileTypeError(file_path)

        result.hostname = match.group('name')
        result.ip = match.group('ip')

        cmd_dict = {}
        content = []
        command = ''

        lines = stream.splitlines()
        if re.match(command_line_reg, lines[0]):
            state = 1  # 当开头是以----开头的时候，是情况一

        else:
            state = 2  # 否则是情况二

        if state == 1:
            for line in lines:
                if re.match(command_line_reg, line):  # 当遇到的是第一种情况
                    if command:  # 当有命令的时候，说明是上一个命令的结尾，要保存
                        result.add_cmd(command, '\n'.join(content))
                    command = re.match(command_line_reg, line).group('cmd')
                    content.clear()  # 清空内容
                    continue
                content.append(line)

        if state == 2:
            for line in lines:
                if not command:
                    command = line.strip()  # 第一行为命令
                    continue
                if re.match(command_line_reg2, line):
                    result.add_cmd(command, '\n'.join(content))
                    command = ''
                    content.clear()
                    continue
                content.append(line)

        if content:
            cmd_dict[command] = '\n'.join(content)

        result.cmd_dict = cmd_dict
        return result



from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Type
import re

from .func import reg_extend

if TYPE_CHECKING:
    from .domain import Cmd


class DefaultVendor:
    """默认厂商"""

    PLATFORM = 'default'
    VERSION_COMMAND = None  # 确定厂商信息的命令
    KEYWORD_REG = None  # 用于匹配厂商的关键字正则表达式
    INVALID_STR = None  # 如果命令返回的结果中包含该字符串, 则认为命令无效

    @classmethod
    def check_vendor(cls, cmds: Dict[str, Cmd]) -> Type[DefaultVendor]:
        """检查确认设备的厂商"""
        for handler in cls.__subclasses__():
            if handler._check_vendor(cmds):
                return handler

        return cls

    @classmethod
    def _check_vendor(cls, cmds: Dict[str, Cmd]) -> bool:
        """子类用于检查设备的厂商以及平台"""
        version_cmd = cls._check_version_command(cmds)
        if version_cmd:
            return (
                True
                if re.search(cls.KEYWORD_REG, cmds[version_cmd].content, re.IGNORECASE)
                else False
            )

        return False

    @classmethod
    def _check_version_command(cls, cmds: Dict[str, Cmd]) -> str:
        """子类用于检查版本命令"""
        for cmd in cmds.keys():
            if re.match(
                re.compile('(' + reg_extend(cls.VERSION_COMMAND) + ')$', re.IGNORECASE),
                cmd,
            ):
                if cmds[cmd].check_valid(cls.INVALID_STR):
                    return cmd
        return ''


class Huawei(DefaultVendor):

    PLATFORM = 'huawei_vrp'
    VERSION_COMMAND = 'dis[[play]] ver[[sion]]'
    KEYWORD_REG = r'Huawei Versatile Routing Platform'
    INVALID_STR = r'Error:.*found at|% .* position'


class H3C(DefaultVendor):

    PLATFORM = 'hp_comware'  # 新华三在海外用的是 HP Comware
    VERSION_COMMAND = 'dis[[play]] ver[[sion]]'
    KEYWORD_REG = r'H3C Comware Platform Software|H3C Comware Software'
    INVALID_STR = r'Ambiguous command found at|% .* found at'


class Cisco(DefaultVendor):

    PLATFORM = 'cisco_ios'
    VERSION_COMMAND = 'sh[[ow]] ver[[sion]]'
    KEYWORD_REG = r'Cisco IOS Software|Cisco Internetwork Operating System Software'
    INVALID_STR = r'Invalid input detected'


class Maipu(DefaultVendor):

    PLATFORM = 'maipu_mypower'
    VERSION_COMMAND = 'sh[[ow]] ver[[sion]]'
    KEYWORD_REG = r'MyPower \(R\) Operating System Software'
    INVALID_STR = r'Invalid input detected at|% Unknown command'


class Ruijie(DefaultVendor):

    PLATFORM = 'ruijie_os'
    VERSION_COMMAND = 'sh[[ow]] ver[[sion]]'
    KEYWORD_REG = r'Ruijie Networks'
    INVALID_STR = r'Invalid input detected at'



from __future__ import annotations

from typing import TYPE_CHECKING

from .. import vendor
from ..analysis_plugin import AnalysisPluginAbc, analysis

if TYPE_CHECKING:
    from ..analysis_plugin import TemplateInfo
    from ..domain import AnalysisResult


class AnalysisPluginWithCpuStatus(AnalysisPluginAbc):
    """
    检查CPU利用率不能过高。
    """

    @analysis.vendor(vendor.Huawei)
    @analysis.template_key('huawei_vrp_display_cpu-usage.textfsm', ['cpu_5_min'])
    def huawei_vrp(template_info: TemplateInfo, result: AnalysisResult):
        """
        CPU利用率高于80%，则告警。
        """
        for row in template_info['display cpu-usage']:
            check_cpu_usage(float(row['cpu_5_min']), result)

    @analysis.vendor(vendor.Cisco)
    @analysis.template_key('cisco_ios_show_processes_cpu.textfsm', ['cpu_5_min'])
    def cisco_ios(template_info: TemplateInfo, result: AnalysisResult):
        """
        CPU利用率高于80%，则告警。
        """
        for row in template_info['show processes cpu']:
            check_cpu_usage(float(row['cpu_5_min']), result)

    @analysis.vendor(vendor.Ruijie)
    @analysis.template_key('ruijie_os_show_cpu.textfsm', ['cpu_5_min'])
    def ruijie_os(template_info: TemplateInfo, result: AnalysisResult):
        """
        CPU利用率高于80%，则告警。
        """
        for row in template_info['show cpu']:
            check_cpu_usage(float(row['cpu_5_min']), result)

    @analysis.vendor(vendor.Maipu)
    @analysis.template_key('maipu_mypower_show_cpu_monitor.textfsm', ['cpu_5_min'])
    def maipu_mypower(template_info: TemplateInfo, result: AnalysisResult):
        """
        CPU利用率高于80%，则告警。
        """
        for row in template_info['show cpu monitor']:
            check_cpu_usage(float(row['cpu_5_min']), result)

    @analysis.vendor(vendor.H3C)
    @analysis.template_key('hp_comware_display_cpu-usage.textfsm', ['cpu_5_min'])
    def hp_comware(template_info: TemplateInfo, result: AnalysisResult):
        """
        CPU利用率高于80%，则告警。
        """
        for row in template_info['display cpu-usage']:
            check_cpu_usage(float(row['cpu_5_min']), result)


def check_cpu_usage(cpu_usage: float, result: AnalysisResult):
    if cpu_usage > 80:
        result.add_warning(f'CPU利用率达到{cpu_usage}%')



from __future__ import annotations

import abc
import re
from dataclasses import dataclass
from typing import Dict, Iterator, List, Optional, Tuple, Type

from . import exception
from .base_info import BaseInfo, EachVendorDeviceInfo
from .data import pystr
from .func import NoneSkip, StoreFunc, pascal_case_to_snake_case, CaseInsensitiveDict
from .logger import logger
from .vendor import DefaultVendor


class Cluster:
    """作为设备的集合"""

    attributes: Tuple[str, ...] = ['plugin_manager', 'devices', 'base_info_handler']

    def __init__(self):
        self._plugin_manager: PluginManagerAbc = None
        self.devices: DeviceList[Device] = DeviceList()
        self.base_info_handler = EachVendorDeviceInfo()

    def parse(self):
        """递归对每个设备的命令进行解析"""
        logger.info('start parse')
        self.devices.parse(base_info_handler=self.base_info_handler)
        logger.info('parse finished')

    def analysis(self):
        """递归对每个设备进行分析"""
        logger.info('start analysis')
        self.devices.analysis(base_info_handler=self.base_info_handler)
        logger.info('analysis finished')

    @property
    def plugin_manger(self) -> PluginManagerAbc:
        return self._plugin_manager

    @plugin_manger.setter
    def plugin_manager(self, obj: PluginManagerAbc):
        self._plugin_manager = obj

    def search(self, device_name: str) -> List[Device]:
        """搜索设备

        Args:
            device_name: 设备名

        Returns:
            List[Device]: 设备列表
        """
        return self.devices.search(device_name)

    def input_dir(self, dir_path: str, expend: str | List[str] = None):
        """输入整个目录，对目录中的文件进行提取设备和命令, 并保存到self.devices中

        Args:
            dir_path: 目录路径
            expend: 文件扩展名
        """

        logger.info(f'input dir: {dir_path!r}')
        devices_list = self.plugin_manager.input_dir(dir_path, expend)

        for cmd_contents_and_device_info in devices_list:
            self.save_device_with_cmds(cmd_contents_and_device_info)

    def input(self, file_path: str):
        """输入文件，对文件中的设备和命令进行提取，并保存到self.devices中

        Args:
            file_path: 文件路径
        """

        logger.info(f'input file: {file_path!r}')
        try:
            input_plugin_result = self.plugin_manager.input(file_path)
            self.save_device_with_cmds(input_plugin_result)
        except exception.InputFileTypeError:  # pragma: no cover
            logger.info('文件不符合input_plugin标准，跳过: {}'.format(file_path))

    def save_device_with_cmds(self, input_plugin_result: InputPluginResult):
        """将设备和命令保存到self.devices中

        Args:
            cmd_contents_and_deviceinfo: 命令内容和设备信息
        """

        if not input_plugin_result.hostname:  # 如果没有设备名，直接跳过
            return

        device_cls = Device()
        device_cls.vendor = input_plugin_result.vendor
        device_cls._plugin_manager = self.plugin_manager
        device_cls.save_to_cmds(input_plugin_result.cmd_dict)  # 保存命令信息
        device_cls._device_info = input_plugin_result._device_info  # 保存设备简单信息
        self.devices.append(device_cls)

    def output(self, file_path: str = '', params: Dict[str, str] = {}):
        """输出到文件

        Args:
            file_path: 文件路径
            params: 传入output_plugin的参数
        """
        self.plugin_manager._output_plugin.run(self.devices, file_path, params)

    def add_device_with_raw_data(
        self, hostname: str, ip: str, cmd_contents: Dict[str, str]
    ):
        """添加设备和命令

        Args:
            hostname: 设备名
            ip: 设备ip
            cmd_contents: 命令内容
        """
        result = InputPluginResult()
        result.hostname = hostname
        result.ip = ip
        result.cmd_dict = cmd_contents

        self.add_device_use_input_plugin_result(result)

    def add_device_use_input_plugin_result(
        self, input_plugin_result: InputPluginResult
    ):
        """通过 ``InputPluginResult`` 添加设备

        Args:
            input_plugin_result: ``InputPluginResult`` 对象
        """
        self.save_device_with_cmds(input_plugin_result)


class DeviceList(list):
    """设备列表"""

    def __init__(self):
        self._devices: List[Device] = []

    def __getitem__(self, index: int) -> Device:
        return self._devices[index]

    def __iter__(self) -> Iterator[Device]:
        return self._devices.__iter__()

    def __len__(self) -> int:
        return len(self._devices)

    def append(self, device: Device):
        """添加设备

        Args:
            device: 设备
        """
        self._devices.append(device)

    def parse(self, base_info_handler: EachVendorDeviceInfo):
        """递归对每个设备的命令进行解析"""
        for device in self._devices:

            # 当没有设备厂商时，只配置通用信息
            if device.vendor == DefaultVendor:
                logger.debug(
                    f'{pystr.parse_plugin_prefix} device:{device._device_info.name!r} 没有匹配到厂商, 跳过.'
                )
                device.info = base_info_handler.run_general_information(device)
                continue

            device.parse()
            # 将分析到的基础信息放到Device.info中
            device.info = base_info_handler.run_baseinfo_func(device)

    def analysis(self, base_info_handler: EachVendorDeviceInfo):
        """递归对每个设备进行分析, 必须在parse之后执行"""
        for device in self._devices:
            device.analysis()
            # 将分析到的状态信息放到Device.info.analysis中
            base_info_handler.run_analysis_info(device)

    def search(self, device_name: str) -> List[Device]:
        """查找设备

        Args:
            device_name: 设备名

        Returns:
            List[Device]: 设备列表
        """
        return [
            device
            for device in self._devices
            if device_name in device._device_info.name
        ]


@dataclass
class DeviceInfo:
    """用于InputPlugin中获取到的设备信息"""

    name: str = ''  # 设备名
    ip: str = ''  # manager_ip
    file_path: str = ''  # 文件路径


class Device:
    """设备类，用于保存每台设备的信息，并完成对设备的综合解析和查询工作"""

    def __init__(self):
        self.cmds: Dict[str, Cmd] = {}
        self._vendor = DefaultVendor
        self._plugin_manager: PluginManagerAbc = None
        self._device_info: DeviceInfo = None
        self._base_info: BaseInfo = None

        self._analysis_result = AnalysisResult()

    @property
    def info(self) -> BaseInfo:
        """设备基础信息，会在parse和analysis后更新

        Returns:
            BaseInfo: 设备基础信息
        """
        if self._base_info is None:
            return BaseInfo()
        return self._base_info

    @info.setter
    def info(self, obj: BaseInfo):
        """设置设备基础信息"""
        self._base_info = obj

    @property
    def analysis_result(self) -> AnalysisResult:
        return self._analysis_result

    def parse_result(self, cmd: str) -> List[dict]:
        """获取解析命令结果, 参数 **cmd** 会自动进行模糊匹配。

        Args:
            cmd: 需要获取解析结果的命令, 支持模糊匹配

        Returns:
            List[dict] | None: 解析结果
        """
        command = self.search_cmd(cmd)
        if not command:
            return []

        return command.parse_result

    def save_to_cmds(self, cmd_contents: Dict[str, str]):
        """将分割好的命令字典保存到设备的命令列表中

        Args:
            cmd_contents: 命令字典
        """
        for command, content in cmd_contents.items():
            cmd = Cmd(command)
            cmd.content = content
            self.cmds[command] = cmd

        if self.vendor is DefaultVendor:  # 最后再检查厂商
            self.check_vendor()

    def check_vendor(self):
        """检查厂商, 无返回值，会直接修改设备的厂商属性"""
        vendor = self.vendor.check_vendor(self.cmds)
        self._vendor = vendor

    @property
    def vendor(self) -> Type[DefaultVendor]:
        """返回厂商类"""
        return self._vendor

    @vendor.setter
    def vendor(self, vendor: Type[DefaultVendor]):
        """设置厂商类"""
        if not issubclass(vendor, DefaultVendor):
            raise TypeError('vendor must be subclass of DefaultVendor')
        self._vendor = vendor

    def parse(self):
        """对每条cmd进行解析"""
        for _, cmd in self.cmds.items():
            try:
                # 首先判断是否为无效命令
                if not cmd.check_valid(self._vendor.INVALID_STR):
                    raise exception.TemplateError(
                        f'platform: {self._vendor.PLATFORM!r} cmd: {cmd.command!r} 无效命令回显.'
                    )

                parse_result = StoreFunc(
                    self._plugin_manager.parse, cmd, self.vendor.PLATFORM
                )
                cmd.update_parse_reslut(parse_result)
            except exception.TemplateError as e:
                logger.debug(
                    f'{pystr.parse_plugin_prefix} device:{self._device_info.name!r} {str(e)}'
                )

    def analysis(self):
        """对设备进行分析, 需要在parse之后"""
        res = self._plugin_manager.analysis(self)
        self._analysis_result.merge(res)

    def search_cmd(self, cmd_name: str) -> Cmd | NoneSkip:
        """
        查找命令, 返回Cmd类

        Args:
            cmd_name: 命令名

        Returns:
            Cmd | None: 命令类或者是个可以执行for的None类
        """
        res = None  # type: Tuple[Cmd, int]
        cmd_name_split = cmd_name.split()
        cmd_name_len = len(cmd_name_split)

        for command in self.cmds.keys():

            score = 0  # 匹配得分
            is_match = True  # 是否匹配的标签

            # 切分命令
            command_split = command.split()
            if not len(command_split) == cmd_name_len:  # 当命令长度不一致就跳过
                continue

            # 对每个单词进行匹配并打分
            for i, each_command in enumerate(command_split):
                if is_match == False:
                    break
                # 排序长命令与短的命令
                if len(each_command) > len(cmd_name_split[i]):
                    long_each_cmd, short_each_cmd = each_command, cmd_name_split[i]
                else:
                    long_each_cmd, short_each_cmd = cmd_name_split[i], each_command

                # 匹配单词
                if not long_each_cmd.startswith(short_each_cmd):
                    is_match = False
                    break
                score += len(short_each_cmd)

            if is_match:
                # 比较得分, 取最高的
                if (not res) or score > res[1]:
                    res = (self.cmds[command], score)

        return res[0] if res else NoneSkip()


class Cmd:
    """用于存放命令名称，以及命令的内容，等待后续的解析
    在解析完成后删除命令的内容,并存放应有的解析内容"""

    def __init__(self, cmd: str, content: str = ''):
        """
        Args:
            cmd: 命令
            content: 命令回显内容
        """
        self._command: str = ''
        self._content: str = ''
        self._parse_result: List[Dict[str, str]] | StoreFunc = []

        self.command = cmd
        self.content = content

    def __enter__(self):
        return self

    def __exit__(self, type, value, tace):
        pass

    @property
    def command(self) -> str:
        """返回命令字符串"""
        return self._command

    @command.setter
    def command(self, cmd: str):
        """设置命令名称
        自动合并多个空格"""
        self._command = ' '.join(cmd.split())

    @property
    def content(self) -> str:
        """返回回显字符串"""
        return self._content

    @content.setter
    def content(self, stream: str):
        self._content = stream

    def update_parse_reslut(self, result: List[Dict[str, str]] | StoreFunc):
        """在取到解析结果后，更新解析结果, 也可以传入一个 StoreFunc, 作为延迟执行

        Args:
            result: 解析结果
        """
        self._parse_result = result

    def check_valid(self, invalid_str: Optional[str] = None) -> bool:
        """判断命令的内容是否为有效内容
        Args:
            invalid_str: 如果命令的内容包含这个字符串，则认为命令无效

        Returns:
            bool: 命令是否有效
        """
        if bool(self.content.strip()) == False:  # 如果为空则认为无效
            return False
        elif invalid_str and re.search(invalid_str, self.content):  # 如果包含无效字符串则认为无效
            return False
        return True

    @property
    def parse_result(self) -> List[Dict[str, str]]:
        """返回解析结果, 如果是 StoreFunc 则执行并返回结果

        Returns:
            List[Dict[str, str]]: 解析结果
        """
        if isinstance(self._parse_result, StoreFunc):
            try:
                res = []
                tmp_dict = self._parse_result()
                for dict in tmp_dict:
                    res.append(CaseInsensitiveDict(dict))
                self._parse_result = res

            except exception.TemplateError as e:
                self._parse_result = []
                logger.debug(e)

        return self._parse_result


class PluginManagerAbc(abc.ABC):
    """对插件的管理的抽象接口, 管理output和parse"""

    def __init__(
        self,
        input_plugin: Optional[Type[InputPluginAbstract]] = None,
        output_plugin: Optional[Type[OutputPluginAbstract]] = None,
        parse_plugin: Optional[Type[ParsePluginAbstract]] = None,
        analysis_plugin: List[Type[AnalysisPluginAbstract]] = [],
    ):
        self._input_plugin: Optional[InputPluginAbstract] = None
        self._output_plugin: Optional[OutputPluginAbstract] = None
        self._parse_plugin: Optional[ParsePluginAbstract] = None
        self._analysis_plugin: List[AnalysisPluginAbstract] = []

        if input_plugin:
            self.input_plugin = input_plugin
        if output_plugin:
            self.output_plugin = output_plugin
        if parse_plugin:
            self.parse_plugin = parse_plugin
        if analysis_plugin:
            self.analysis_plugin = analysis_plugin

    def check_cls(check_type: str):
        """装饰器用来检测插件的类型, 如果插件不是指定的类型, 则抛出异常
        如果是传入的是实例，改为类再传入
        """

        def wrapper(func):
            def inner(self, plugin_cls: Type[PluginAbstract]):
                if not hasattr(plugin_cls, '__name__'):  # 如果是实例
                    plugin_cls = plugin_cls.__class__
                # 如果不是指定的类型
                if not issubclass(plugin_cls, globals()[check_type]):
                    raise TypeError(f'{plugin_cls.__name__} is not {check_type}')
                return func(self, plugin_cls)

            return inner

        return wrapper

    def check_cls_list(check_type: str):
        """装饰器用来检测插件的类型, 如果插件不是指定的类型, 则抛出异常
        如果是传入的是实例，改为类再传入
        """

        def wrapper(func):
            def inner(self, plugin_cls_list: List[Type[PluginAbstract]]):
                cls_list = []
                for plugin_cls in plugin_cls_list:
                    if not hasattr(plugin_cls, '__name__'):  # 如果是实例
                        plugin_cls = plugin_cls.__class__
                    # 如果不是指定的类型
                    if not issubclass(plugin_cls, globals()[check_type]):
                        raise TypeError(f'{plugin_cls.__name__} is not {check_type}')
                    cls_list.append(plugin_cls)
                return func(self, cls_list)

            return inner

        return wrapper

    @property
    def input_plugin(self) -> Optional[InputPluginAbstract]:
        return self._input_plugin

    @input_plugin.setter
    @check_cls("InputPluginAbstract")
    def input_plugin(self, plugin_cls: Type[InputPluginAbstract]):
        self._input_plugin = plugin_cls()

    @property
    def output_plugin(self) -> Optional[OutputPluginAbstract]:
        return self._output_plugin

    @output_plugin.setter
    @check_cls("OutputPluginAbstract")
    def output_plugin(self, plugin_cls: Type[OutputPluginAbstract]):
        self._output_plugin = plugin_cls()

    @property
    def parse_plugin(self) -> Optional[ParsePluginAbstract]:
        return self._parse_plugin

    @parse_plugin.setter
    @check_cls("ParsePluginAbstract")
    def parse_plugin(self, plugin_cls: Type[ParsePluginAbstract]):
        self._parse_plugin = plugin_cls()

    @property
    def analysis_plugin(self) -> Optional[AnalysisPluginAbstract]:
        return self._analysis_plugin

    @analysis_plugin.setter
    @check_cls_list("AnalysisPluginAbstract")
    def analysis_plugin(self, plugin_cls_list: List[Type[AnalysisPluginAbstract]]):
        self._analysis_plugin = [plugin_cls() for plugin_cls in plugin_cls_list]

    def parse(self, cmd: Cmd, platform: str) -> Dict[str, str]:
        """对单个命令的内容进行解析"""
        if self._parse_plugin is None:
            raise exception.PluginNotSpecify('parse plugin is None')
        return self._parse_plugin.run(cmd, platform)

    def analysis(self, device: Device) -> AnalysisResult:
        """对设备进行分析, 返回分析结果列表"""
        res = AnalysisResult()
        if not self._analysis_plugin:
            raise exception.PluginNotSpecify('analysis plugin list is empty')
        for plugin in self._analysis_plugin:
            res.merge(plugin.run(device))

        return res

    def input(self, file_path: str) -> InputPluginResult:
        """对单个文件进行设备输入"""
        if self._input_plugin is None:
            raise exception.PluginNotSpecify('input plugin is None')
        return self._input_plugin.run(file_path)

    def output(self, devices: DeviceList, file_path: str):
        """对设备列表进行输出"""
        if self._output_plugin is None:
            raise exception.PluginNotSpecify('output plugin is None')
        self._output_plugin.run(devices, file_path)

    @abc.abstractmethod
    def input_dir(
        self, dir_path: str, expend: str | List = None
    ) -> List[InputPluginResult]:
        """对目录中的文件进行设备输入"""
        raise NotImplementedError


class InputPluginResult:
    """输入插件的结果"""

    def __init__(self):
        self._device_info: DeviceInfo = DeviceInfo()
        self._cmd_dict: Dict[str, str] = {}
        self.vendor: DefaultVendor = DefaultVendor

    @property
    def hostname(self):
        """设备名称"""
        return self._device_info.name

    @hostname.setter
    def hostname(self, value: str):
        self._device_info.name = value

    @property
    def ip(self):
        """管理IP"""
        return self._device_info.ip

    @ip.setter
    def ip(self, value: str):
        self._device_info.ip = value

    def add_cmd(self, cmd: str, content: str):
        """添加命令和对应的回显，所有命令将转为小写,
        如果命令已经存在，则取长的, 如果长度相等，取最新的。

        Args:
            cmd: 待添加的命令
            content: 命令对应的回显
        """
        cmd = cmd.strip().lower()
        if cmd in self._cmd_dict and len(self._cmd_dict[cmd]) > len(content):
            return
        self._cmd_dict[cmd] = content

    @property
    def cmd_dict(self) -> Dict[str, str]:
        """返回命令字典"""
        return self._cmd_dict

    @cmd_dict.setter
    def cmd_dict(self, cmd_dict: Dict[str, str]):
        """直接添加命令字典"""
        for cmd, content in cmd_dict.items():
            self.add_cmd(cmd, content)


class AlarmLevel:
    """分析模块执行后，会对告警提供告警级别，并且记录告警内容

    告警分为3级:

    * ``0`` 正常消息，无告警
    * ``1`` 需要关注的告警
    * ``2`` 需要立即处理的告警
    """

    NORMAL = 0
    FOCUS = 1
    WARNING = 2

    def __init__(
        self,
        level: int,
        message: str = '',
        plugin_cls: Type[AnalysisPluginAbstract] = None,
    ):
        self._level = level
        self.message = message
        self.plugin_cls = plugin_cls

    @property
    def level(self) -> int:
        """级别"""
        return self._level

    @level.setter
    def level(self, level: int):
        if level < AlarmLevel.NORMAL or level > AlarmLevel.WARNING:
            raise exception.AnalysisLevelError
        self._level = level

    @property
    def plugin_name(self) -> str:
        """分析插件的类名"""
        if self.plugin_cls is None:
            return ''
        return self.plugin_cls.__name__

    @property
    def is_warning(self) -> bool:
        """是否为警告级别"""
        return self._level == AlarmLevel.WARNING

    @property
    def is_focus(self) -> bool:
        """是否为关注级别"""
        return self._level == AlarmLevel.FOCUS

    @property
    def is_normal(self) -> bool:
        """是否为正常级别"""
        return self._level == AlarmLevel.NORMAL

    @property
    def above_focus(self) -> bool:
        """是否在关注级别及以上"""
        return self._level >= AlarmLevel.FOCUS

    @property
    def level_str(self) -> str:
        """级别描述"""
        if self._level == AlarmLevel.NORMAL:
            return 'NORMAL'
        elif self._level == AlarmLevel.FOCUS:
            return 'FOCUS'
        elif self._level == AlarmLevel.WARNING:
            return 'WARNING'

    @property
    def doc(self) -> str:
        """返回分析模块的文档"""
        return self.plugin_cls.__doc__.strip() if self.plugin_cls else ''


class AnalysisResult:
    """分析结果"""

    def __init__(self):
        self._result: List[AlarmLevel] = []

    def merge(self, result: AnalysisResult):
        """合并分析结果"""
        self._result.extend(result._result)

    def add(self, level: AlarmLevel):
        """添加分析结果"""
        self._result.append(level)

    def add_normal(self, message: str = ''):
        """添加正常结果"""
        self.add(AlarmLevel(AlarmLevel.NORMAL, message))

    def add_focus(self, message: str = ''):
        """添加关注结果"""
        self.add(AlarmLevel(AlarmLevel.FOCUS, message))

    def add_warning(self, message: str = ''):
        """添加警告结果"""
        self.add(AlarmLevel(AlarmLevel.WARNING, message))

    def get(self, plugin_name: str) -> AnalysisResult:
        """
        获取指定插件的结果

        ``plugin_name`` 可以写的形式：

        * 完整插件名称 (e.g ``AnalysisPluginWithPowerStatus`` )
        * 下划线 (e.g ``analysis_plugin_with_power_status`` )
        * 全小写 (e.g ``analysispluginwithpowerstatus`` )
        * 简写 (e.g ``power status`` 或者 ``power_status'`` )

        Args:
            plugin_name: 插件名称

        Return:
            AlarmLevel: AlarmLevel对象
        """
        ret = AnalysisResult()
        for alarm in self._result:
            if self._short(alarm.plugin_name) == self._short(plugin_name):
                ret.add(alarm)
        return ret

    def _short(self, plugin_name: str) -> str:
        """获取指定插件的简写
        e.g: AnalysisPluginWithPower -> power
        """
        name = plugin_name.lower()
        name = name.replace(' ', '')
        name = name.replace('_', '')
        name = name.replace('analysispluginwith', '')
        return name

    def __getitem__(self, index) -> AlarmLevel:
        return self._result[index]

    def __iter__(self) -> Iterator[AlarmLevel]:
        return iter(self._result)

    def __len__(self) -> int:
        return len(self._result)

    @property
    def include_warning(self) -> bool:
        """是否包含警告级别"""
        for alarm in self._result:
            if alarm.is_warning:
                return True
        return False


class PluginAbstract(abc.ABC):
    """插件的抽象"""

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError

    @property
    def doc(self) -> str:
        """返回分析模块的文档"""
        return self.__doc__.strip() if self.__doc__ else ''

    @property
    def short_name(self) -> str:
        """返回模块的简写"""
        return pascal_case_to_snake_case(self.__class__.__name__.split('With', 1)[-1])


class InputPluginAbstract(PluginAbstract):
    def run(self, file_path: str) -> InputPluginResult:
        """对单个文件进行设备输入

        Args:
            file_path: 文件路径

        Returens:
            InputPluginResult: 输入插件的返回结果
        """

        with open(file_path, 'r', encoding='utf_8_sig', errors='ignore') as f:
            stream = f.read()
        result = self.main(file_path, stream)
        result._device_info.file_path = file_path
        return result

    @abc.abstractmethod
    def main(self, file_path: str, stream: str) -> InputPluginResult:
        """重载此方法进行输入

        Args:
            file_path: 文件路径
            stream: 文件内容

        Returns:
            InputPluginResult: 输入插件的返回结果
        """

        raise NotImplementedError


class OutputPluginAbstract(PluginAbstract):
    @dataclass
    class OutputArgs:

        devices: DeviceList  # 设备列表
        file_path: str  # 输出文件的路径
        output_params: Dict[str, str]  # 输出文件的参数

    def run(
        self,
        devices: DeviceList[Device],
        path: str,
        output_params: Optional[Dict[str, str]],
    ):
        """对设备列表进行输出

        Args:
            devices: 设备列表
            path: 输出文件的路径
            output_params: 输出文件的参数
        """

        self.args = self.OutputArgs(
            devices=devices, file_path=path, output_params=output_params
        )
        return self.main()

    @abc.abstractmethod
    def main(self):
        """重载此方法进行输出"""
        raise NotImplementedError

    def check_args(self, *args: str):
        """检查必须要提供的参数都已经存在, 否则抛出异常

        Args:
            *args: 期望需要的参数

        Raises:
            exception.OutputPluginArgsError: 参数不足
        """
        for arg in args:
            if arg not in self.args.output_params:
                raise exception.OutputParamsNotGiven(self.__class__.__name__, arg)


class ParsePluginAbstract(PluginAbstract):
    def run(self, cmd: Cmd, platform: str) -> Dict[str, str]:
        return self.main(cmd, platform)

    @abc.abstractmethod
    def main(self, cmd: Cmd, platform: str) -> Dict[str, str]:
        raise NotImplementedError

    @abc.abstractmethod
    def set_external_templates(self, template_dir: str):
        raise NotImplementedError


class AnalysisPluginAbstract(PluginAbstract):
    @abc.abstractmethod
    def run(self, device: Device) -> AnalysisResult:
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, Type

import rich_typer as typer
from rich import print
from rich.console import Console
from rich.table import Table
from rich.panel import Panel

from . import NetInspect, __version__, exception
from .analysis_plugin import analysis
from .func import print_err

if TYPE_CHECKING:
    from .domain import PluginAbstract
    from plugins.parse_plugin_with_ntc_templates import ParsePluginWithNtcTemplates
    from textfsm.clitable import CliTable


app = typer.RichTyper(add_completion=False)
console = Console()

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
BANNER = (
    f"[b]Network Inspect[/b] [magenta]v{__version__}[/] 🤑\n\n[dim]网络设备数据结构化分析框架[/]\n"
)
URL = "♥ https://github.com/Elinpf/net_inspect"


def print_table(table_name: str, items: list, rows: list):
    """打印表格"""
    table = Table(title=f"[bold reverse cyan]{table_name}[/]", title_justify='left')
    for item in items:
        table.add_column(item)
    for row in rows:
        table.add_row(*row)

    table.expand = True
    console.print(table, justify='left')


def print_panel(title: str, items: list, rows: list):
    table = Table(highlight=True, box=None, show_header=False)
    for item in items:
        table.add_column(item)
    for row in rows:
        table.add_row(*row)

    panel = Panel(table, border_style="dim", title=title, title_align="left")
    console.print(panel)


def print_plugin_list(all_plugins: Dict[str, Dict[str, Type[PluginAbstract]]]):
    """打印插件列表"""
    column_names = ['插件名称', '简写', '描述']
    for plugin_type, plugins in all_plugins.items():
        row_list = []
        for plugin_name, plugin_cls in plugins.items():
            instance = plugin_cls()
            row_list.append([plugin_name, instance.short_name, instance.doc])

        print_table(plugin_type, column_names, row_list)
        print()


def print_analysis_list():
    """打印每个分析模块支持的厂商"""
    analysis_vendor_dict = {}
    for info in analysis.store:
        if info.plugin_name not in analysis_vendor_dict:
            analysis_vendor_dict[info.plugin_name] = []
        analysis_vendor_dict[info.plugin_name].append(info.vendor_platform)

    # 对厂商进行排序
    rows = []
    for plugin_name, vendors in analysis_vendor_dict.items():
        rows.append([plugin_name, ', '.join(sorted(vendors)), str(len(vendors))])

    # 打印表格
    print_table('分析模块支持的厂商平台', ['插件名称', '支持平台', '数量'], rows)
    print()


def print_base_info_list():
    """打印基础信息所包含的属性"""
    from .base_info import AnalysisInfo, BaseInfo

    base_info_field = BaseInfo().__annotations__
    analysis_info_field = AnalysisInfo().__annotations__
    rows = []
    for info, type in base_info_field.items():  # base_info中的属性
        if info == 'analysis':
            continue
        rows.append([info, type.replace('[', '\[')])

    for info, type in analysis_info_field.items():  # analysis中的属性
        rows.append(['analysis.' + info, type.replace('[', '\[')])

    print_table('基础信息所包含的属性', ['属性名称', '类型'], rows)
    print()


def print_ntc_template_value(clitable: CliTable, template_name: str):
    """打印ntc_templates中的模板VALUE"""

    values = clitable.header.values

    rows = []
    for value in values:
        rows.append([value.lower()])

    print()
    print_panel(f'[red]{template_name}[/] 中的 [blue]VALUES[/]', ['VALUES'], rows)
    print()


@app.callback(
    invoke_without_command=True,
    banner=BANNER,
    banner_justify='center',
    context_settings=CONTEXT_SETTINGS,
    epilog=URL,
)
def main(
    ctx: typer.Context,
    input_path: str = typer.Option(None, '--input', '-i', help='输入的文件路径，可以是文件夹'),
    input_plugin: str = typer.Option('console', '--input-plugin', '-I', help='设置输入插件'),
    output_plugin: str = typer.Option(
        'device_list', '--output-plugin', '-O', help='设置输出插件'
    ),
    output_path: str = typer.Option('', '--output', '-o', help='输出的文件路径'),
    verbose: int = typer.Option(
        0, '--verbose', '-v', help='设置DEBUG等级(-v, -vv)', count=True, show_default=False
    ),
    plugin_list: bool = typer.Option(False, '--plugin-list', '-l', help='显示插件列表'),
    analysis_list: bool = typer.Option(
        False, '--analysis-list', '-L', help='显示分析插件所支持的平台列表'
    ),
    base_info_list: bool = typer.Option(
        False, '--base-info-list', '-b', help='显示基础信息属性列表'
    ),
):
    net = NetInspect()
    net.set_plugins(input_plugin=input_plugin, output_plugin=output_plugin)
    if verbose == 1:
        net.enable_console_log('INFO')
    elif verbose == 2:
        net.enable_console_log('DEBUG')

    if plugin_list:
        print()
        print_plugin_list(net.get_all_plugins())
        exit()

    if analysis_list:
        print()
        print_analysis_list()
        exit()

    if base_info_list:
        print()
        print_base_info_list()
        exit()

    if input_path:
        net.run(input_path=input_path, output_file_path=output_path)
        exit()


@app.command(
    context_settings=CONTEXT_SETTINGS,
    epilog=URL,
)
def textfsm(
    ctx: typer.Context,
    external_path: str = typer.Option('', '--external-path', '-e', help='设置外部模板路径'),
    platform: str = typer.Option(..., '--platform', '-p', help='厂商平台'),
    command: str = typer.Option(..., '--command', '-c', help='命令'),
):
    """查看模板仓库中模板的 [blue]VALUES[/]"""

    net = NetInspect()

    plg = net.get_parse_plugins()[
        'ParsePluginWithNtcTemplates'
    ]()  # type: ParsePluginWithNtcTemplates

    if external_path:
        plg.set_external_templates(external_path)

    try:
        res = plg.get_clitable(command, platform)
        template_name = "{platform}_{command}.textfsm".format(
            platform=res['platform'], command=res['command'].replace(' ', '_')
        )
        print_ntc_template_value(res['clitable'], template_name)
    except exception.TemplateNotSupperThisPlatform as e:
        print_err(f'模板仓库不支持此平台: {e.platform}')

    except exception.TemplateNotSupperThisCommand as e:
        print_err(f"模板仓库中没有找到此命令: {e.command!r}")



from __future__ import annotations

import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Dict, Iterator, List, Tuple, Type

from . import exception
from .domain import AlarmLevel, AnalysisPluginAbstract, AnalysisResult
from .func import get_command_from_textfsm, snake_case_to_pascal_case, Singleton

if TYPE_CHECKING:
    from .domain import DefaultVendor, Device

TEMPLATE = str  # ntc-templates 模板文件名称
KEY = str  # ntc-templates 模板文件中的变量名称
PLUGIN_NAME = str  # 分析模块的类名称


@dataclass(unsafe_hash=True)
class AnalysisFunctionInfo:
    plugin_name: str  # 分析插件的名称
    vendor_platform: str  # 分析模块的类名称
    vendor: Type[DefaultVendor]  # 厂商的类
    function_name: str  # 分析函数名称
    function: Callable[[Dict[TEMPLATE, List[KEY]], AnalysisResult]]  # 存放执行的分析函数
    template_keys_list: List[Tuple[TEMPLATE, List[KEY]]]  # 模板文件中的变量名称列表
    template_keys_value: TemplateKeyValue  # 存放模板文件中的变量名称和变量值
    base_info_keys_list: List[str]  # 存放基本信息的字段列表
    doc: str  # 分析函数的注释文档


class StoreTemplateKey(Singleton):
    """
    用于存放分析插件的支持厂商和需要模板的值
    这个实例时独立于分析插件的实例，分析插件的run方法会调用这个实例对应Plugin和Vendor的分析函数

    这个类为全局单例类，用于存放所有分析插件的信息
    """

    def __init__(self):
        self.store: List[AnalysisFunctionInfo] = []
        # 临时存储 template 和 base_info
        self._temp_store: Dict[str, List[Tuple[str] | str]] = {
            'template': [],
            'base_info': [],
        }
        self._only_run_plugins: List[str] = []  # 只运行指定的插件

    def _clear_temp_store(self):
        """清空临时存储"""

        self._temp_store = {
            'template': [],
            'base_info': [],
        }

    def set_only_run_plugins(self, plugins: List[Type[AnalysisPluginAbstract]]):
        """
        设置只运行指定的插件

        Args:
            plugins: 指定的插件列表
        """
        self._only_run_plugins = [plugin.__name__ for plugin in plugins]

    def get_funcs(
        self, plugin_name: PLUGIN_NAME, vendor: Type[DefaultVendor]
    ) -> Iterator[Callable, TemplateKeyValue]:
        """
        获取指定厂商的分析函数, 并且返回模板名称和变量名称的迭代器

        如果设置了only_run_plugins，则只运行指定的插件

        Args:
            plugin_name: 分析插件类名称
            vendor: 厂商

        Return:
            分析函数和TemplateKey的迭代器
        """
        vendor_name = vendor.PLATFORM
        func_list = []

        for info in self.filter(plugin_name=plugin_name, vendor_platform=vendor_name):
            if (
                self._only_run_plugins and plugin_name not in self._only_run_plugins
            ):  # 如果设置了only_run_plugins，则只运行指定的插件
                continue
            func_list.append(
                (info.function, info.template_keys_value, info.base_info_keys_list)
            )

        for func, temp_keys, base_info_keys_list in func_list:
            yield func, temp_keys, base_info_keys_list

    def filter(
        self,
        plugin_name: PLUGIN_NAME = None,
        vendor_platform: str = None,
        function_name: str = None,
    ) -> List[AnalysisFunctionInfo]:
        """
        返回指定的分析函数列表

        Args:
            plugin_name: 分析插件类名称
            vendor_platform: 厂商平台名称
            function_name: 分析函数名称

        Return:
            分析函数的列表
        """
        res = []
        for info in self.store:

            # 如果名称是以蛇形命名，则转换为驼峰命名
            if '_' in plugin_name:
                plugin_name = snake_case_to_pascal_case(plugin_name)

            if plugin_name and info.plugin_name != plugin_name:
                continue
            if vendor_platform and info.vendor_platform != vendor_platform:
                continue
            if function_name and info.function_name != function_name:
                continue
            res.append(info)
        return res

    def store_vendor(self, vendor: Type[DefaultVendor], func: Callable):
        """
        将存储的模板名称和变量名称放入对应的厂商存储器中

        Args:
            vendor: 厂商类
            func: 厂商的分析函数
        """
        plugin_name, function_name = func.__qualname__.split('.', maxsplit=2)
        template_keys_value = TemplateKeyValue(vendor.PLATFORM)
        for template_name, keys in self._temp_store['template']:
            template_keys_value.append(template_name, keys)

        af = AnalysisFunctionInfo(
            plugin_name=plugin_name,
            vendor_platform=vendor.PLATFORM,
            vendor=vendor,
            function_name=function_name,
            function=func,
            template_keys_list=list(self._temp_store['template']),
            template_keys_value=template_keys_value,
            base_info_keys_list=self._temp_store['base_info'],
            doc=func.__doc__.strip(),
        )

        self.store.append(af)
        self._clear_temp_store()

    def vendor(self, vendor: DefaultVendor):
        """
        方法的装饰器，用来确定装饰的方法是对哪个厂商的分析

        Args:
            vendor: 厂商
        """

        def func_init(func):
            self.store_vendor(vendor, func)
            return func

        return func_init

    def template_key(self, template: TEMPLATE, keys: List[KEY]):
        """
        方法的装饰器，用来确定装饰的方法需要使用哪些模板和内容

        Args:
            template: 模板文件名
            keys: 变量名称列表
        """

        def func_init(func):
            self._temp_store['template'].append((template, keys))
            return func

        return func_init

    def base_info(self, *keys: List[str]):
        """
        将需要用到的BaseInfo字段存储起来，在使用的时候调用

        Args:
            keys: 需要 BaseInfo 中的哪些字段
        """

        def func_init(func):
            self._temp_store['base_info'].extend(keys)
            return func

        return func_init


analysis = StoreTemplateKey()


# FIXME 这个类模糊
class AlarmLevel(AlarmLevel):
    @property
    def level(self):
        return self._level

    @level.setter
    def level(self, level: int):
        """
        设置告警级别

        Args:
            level: 告警级别

        Raises:
            ValueError: 告警级别不在0-2之间
        """
        if level < AlarmLevel.NORMAL or level > AlarmLevel.WARNING:
            raise exception.AnalysisLevelError
        self._level = level


class TemplateKeyValue:
    def __init__(self, vendor_platform: str):
        """
        用于存放ntc-templates模板名称和里面的变量以及变量对应的值

        Args:
            - vendor_platform: 厂商平台的字符串
        """
        self._vendor_platform = vendor_platform  # 平台名称字符串
        self._key_store: Dict[TEMPLATE, Dict[KEY, List[str]]] = {}  # 存储模板的名称和变量对应关系

    def append(self, template: TEMPLATE, keys: List[KEY]):
        """
        增加ntc-template模板名称和对应的变量列表
        """
        self._key_store[template] = keys


class BaseInfoKeyValue(dict):
    ...


@dataclass
class TemplateInfo:
    """
    用于存放只给到各个分析函数的数据
    """

    template_key_value: Dict[TEMPLATE, List[Dict[KEY, str]]]  # 模板变量和值的字典
    vendor_platform: str  # 厂商平台的字符串
    base_info: BaseInfoKeyValue  # BaseInfo的值

    def __getitem__(self, name: TEMPLATE) -> List[Dict[KEY, str]]:
        """
        通过模板名称获取模板变量和值的字典列表
        模板名称可以使用的形式：
        e.g:
        - 完整形式: huawei_vrp_display_version.textfsm
        - 完整命令: display version

        Args:
            name: 模板名称

        Returns:
            模板变量和值的字典列表

        Raises:
            exception.NtcTemplateNotDefined: 使用缩写或者不存在的模板名称时抛出异常
        """
        name = self._from_command_to_template_file(name)
        if not name in self.template_key_value:
            raise exception.NtcTemplateNotDefined(name)
        return self.template_key_value[name]

    def _from_command_to_template_file(self, command: str) -> TEMPLATE:
        """
        将命令还原为模板文件名
        Args:
            command 命令

        Return:
            模板文件名

        >>> self._from_command_to_template_file('display interface status')
        'huawei_os_display_interface_status.textfsm'
        """
        command = command.lower()
        command = command.replace(' ', '_')
        if not command.startswith(self.vendor_platform):
            command = self.vendor_platform + '_' + command

        if not command.endswith('.textfsm'):
            command += '.textfsm'

        return command


class AnalysisPluginAbc(AnalysisPluginAbstract):
    def _get_template_key_value(
        self,
        template_keys: TemplateKeyValue,
        device: Device,
    ) -> Dict[TEMPLATE, List[Dict[KEY, str]]]:
        """
        更新对应模板中的对应变量的值
        搜索Device中的cmd，将需要用到的模板和值取出来放到返回值中

        Args:
            template_keys: 模板和值字典
            device: 设备对象

        Returns:
            模板和值的字典

        Raises:
            exception.AnalysisTemplateNameError  名称后缀不是.textfsm
            KeyError 模板中的Value不存在

        """
        ret = {}

        for template_file, keys in template_keys._key_store.items():
            # 当结尾不是.textfsm时，补充完整
            if not template_file.endswith('.textfsm'):
                template_file += '.textfsm'

            cmd = get_command_from_textfsm(  # 通过模板文件名获得命令
                device.vendor.PLATFORM, template_file
            )
            cmd_find = device.search_cmd(cmd)  # 搜索命令
            if not cmd_find:  # 如果命令不存在，跳过
                ret[template_file] = []  # 并且返回一个空的列表，作为占位符
                continue

            temp_list = []
            for row in cmd_find.parse_result:  # 将需要的键值取出来
                try:
                    _ = {key: row[key.lower()] for key in keys}
                    temp_list.append(_)
                except KeyError as e:  # pragma: no cover
                    raise KeyError(f'{self.__class__.__name__}模板中的键值{str(e)}不存在')

            ret[template_file] = temp_list
        return ret

    def _get_base_info_key_value(
        self, base_info_keys: list, device: Device
    ) -> BaseInfoKeyValue:
        """将需要的BaseInfo的键值取出来"""
        base_info_value = BaseInfoKeyValue()
        for key in base_info_keys:
            if not hasattr(device.info, key):
                raise KeyError(f'{self.__class__.__name__}中的键值{key!r}不在base_info中')

            else:
                base_info_value[key] = getattr(device.info, key)
        return base_info_value

    def run(self, device: Device) -> AnalysisResult:
        try:
            result = AnalysisResult()

            klass_name = self.__class__.__name__
            for func, template_keys, base_info_keys in analysis.get_funcs(
                klass_name, device.vendor
            ):
                # 执行对应厂商的分析方法
                template_key_value = self._get_template_key_value(template_keys, device)
                base_info_value = self._get_base_info_key_value(base_info_keys, device)

                # 没有数据时，说明不存在可以分析的命令，跳过
                skip_flag = True
                for k, v in template_key_value.items():
                    if v:
                        skip_flag = False
                        break

                for k, v in base_info_value.items():
                    if v:
                        skip_flag = False
                        break

                if skip_flag:
                    continue

                # 放到数据类中
                template_info = TemplateInfo(
                    template_key_value=template_key_value,
                    vendor_platform=device.vendor.PLATFORM,
                    base_info=base_info_value,
                )

                each_result = AnalysisResult()

                try:
                    func(template_info, each_result)
                except TypeError as e:  # pragma: no cover
                    msg = "方法只需要两个参数，第一个是TemplateInfo，第二个是AnalysisResult, 不需要self"
                    raise TypeError(f'{klass_name}.{func.__name__}() {msg}') from e

                # 当没有分析结果的时候，说明没有问题，给出一个正常级别提示
                if not each_result._result:
                    result.add_normal()

                # 将分析结果放到总结果中
                result.merge(each_result)

        except exception.NtcTemplateNotDefined as e:  # pragma: no cover
            raise exception.NtcTemplateNotDefined(
                f"{self.__class__.__name__} 中无法识别对应模板 {e}, 请检查分析方法中调用的模板名是否正确。"
            )

        for alarm in iter(result):  # 设置告警所属插件类
            if not alarm.plugin_cls:
                alarm.plugin_cls = self.__class__

        return result



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Dict, List, Optional, Type

from .bootstrap import bootstrap
from .data import pyoption, pystr
from .domain import Cluster
from .logger import LoggerConfig, logger
from .plugin_manager import PluginManager

if TYPE_CHECKING:
    from .base_info import BaseInfo, EachVendorDeviceInfo
    from .domain import (
        Device,
        InputPluginAbstract,
        InputPluginResult,
        OutputPluginAbstract,
        ParsePluginAbstract,
        PluginAbstract,
    )


class NetInspect:
    def __init__(self):
        self._plugins = bootstrap()
        self._plugin_manager = PluginManager()
        self.cluster = Cluster()
        self._logconfig = LoggerConfig()

        self._plugin_manager.parse_plugin = self._plugins.get_parse_plugin(
            pystr.default_parse_plugin
        )
        self._plugin_manager.analysis_plugin = self._plugins.get_analysis_plugin_list()
        self.cluster._plugin_manager = self._plugin_manager

    def enable_console_log(self, level: str = '', log_format: str = ''):
        """启用控制台日志

        Args:
            level: 日志级别
            log_format: 日志格式，可以为空使用默认值
        """
        self._logconfig.enable_console_log(
            level=level or pyoption.console_log_level,
            log_format=log_format or pyoption.console_format,
        )

    def enable_file_log(
        self,
        file_path: str = '',
        level: str = '',
        rotation: str = '',
        log_format: str = '',
    ):
        """启用文件日志

        Args:
            file_path: 日志文件路径, 为空使用默认值
            level: 日志级别, 为空使用默认值
            rotation: 日志轮转, 为空使用默认值
            log_format: 日志格式, 为空使用默认值
        """

        self._logconfig.enable_file_log(
            file_path or pyoption.logfile_name,
            level=level or 'DEBUG',
            rotation=rotation or pyoption.logfile_rotation,
            log_format=log_format or pyoption.logfile_format,
        )

    def get_all_plugins(self) -> Dict[str, Dict[str, Type[PluginAbstract]]]:
        """获取所有插件

        Returns:
            所有插件以字典的形式返回
        """
        return {
            'input_plugins': self.get_input_plugins(),
            'output_plugins': self.get_output_plugins(),
            'parse_plugins': self.get_parse_plugins(),
            'analysis_plugins': self.get_analysis_plugins(),
        }

    def get_input_plugins(self) -> Dict[str, Type[PluginAbstract]]:
        """取得所有的输入插件

        Returns:
            返回所有的输入插件的字典类型
        """
        return self._plugins.input_plugins

    def get_output_plugins(self) -> Dict[str, Type[PluginAbstract]]:
        """取得所有的输出插件

        Returns:
            返回所有的输出插件的字典类型
        """
        return self._plugins.output_plugins

    def get_parse_plugins(self) -> Dict[str, Type[PluginAbstract]]:
        """取得所有的解析插件

        Returns:
            返回所有的解析插件的字典类型
        """
        return self._plugins.parse_plugins

    def get_analysis_plugins(self) -> Dict[str, Type[PluginAbstract]]:
        """取得所有的分析插件"""
        return self._plugins.analysis_plugins

    def set_input_plugin(self, plugin_cls: Type[InputPluginAbstract] | str):
        """设置输入插件

        Args:
            plugin_cls: 插件类或者插件名称
        """
        if type(plugin_cls) == str:
            plugin_cls = self._plugins.get_input_plugin(plugin_cls)
        self._plugin_manager.input_plugin = plugin_cls

    def set_output_plugin(self, plugin_cls: Type[OutputPluginAbstract] | str):
        """设置输出插件

        Args:
            plugin_cls: 插件类或者插件名称
        """
        if type(plugin_cls) == str:
            plugin_cls = self._plugins.get_output_plugin(plugin_cls)
        self._plugin_manager.output_plugin = plugin_cls

    def set_parse_plugin(self, plugin_cls: Type[ParsePluginAbstract] | str):
        """设置解析插件

        Args:
            plugin_cls: 插件类或者插件名称
        """
        if type(plugin_cls) == str:
            plugin_cls = self._plugins.get_parse_plugin(plugin_cls)
        self._plugin_manager.parse_plugin = plugin_cls

    def set_plugins(
        self,
        input_plugin: Optional[Type[InputPluginAbstract] | str] = None,
        output_plugin: Optional[Type[OutputPluginAbstract] | str] = None,
        parse_plugin: Optional[Type[ParsePluginAbstract] | str] = None,
    ):
        """设置插件

        Args:
            input_plugin: 输入插件
            output_plugin: 输出插件
            parse_plugin: 解析插件
        """
        if input_plugin:
            self.set_input_plugin(input_plugin)
        if output_plugin:
            self.set_output_plugin(output_plugin)
        if parse_plugin:
            self.set_parse_plugin(parse_plugin)

    def set_external_templates(self, templates_dir: str):
        """设置外部模板目录

        Args:
            templates_dir: 模板目录路径
        """
        self._plugin_manager.parse_plugin.set_external_templates(templates_dir)

    @logger.catch(reraise=True)
    def run_input(self, path: str) -> Cluster:
        """运行输入插件, 如果没有指定输入插件则跳过

        Args:
            path: 输入路径

        Returns:
            Cluster: 集群对象
        """

        if not self._plugin_manager.input_plugin:
            logger.info('未指定`input_plugin`, 跳过 `run_input` 函数.')
            return self.cluster

        if os.path.isfile(path):
            self.cluster.input(path)
        elif os.path.isdir(path):
            self.cluster.input_dir(path)
        else:
            raise ValueError('`path`必须是文件或者目录')

        logger.info('输入插件运行完成, 总共发现 {} 台设备.', len(self.cluster.devices))
        return self.cluster

    @logger.catch(reraise=True)
    def run_parse(self) -> Cluster:
        """运行解析插件

        Returns:
            Cluster: 解析后的集群
        """
        self.cluster.parse()
        return self.cluster

    @logger.catch(reraise=True)
    def run_analysis(self) -> Cluster:
        """运行分析插件

        Returns:
            Cluster: 分析后的集群
        """
        self.cluster.analysis()
        return self.cluster

    @logger.catch(reraise=True)
    def run_output(self, file_path: str = '', params: Dict[str, str] = {}):
        """运行输出插件

        Args:
            file_path: 输出文件路径, 为空使用默认值
            params: 传递给output_plugin的参数
        """
        if self._plugin_manager.output_plugin:  # 当有output插件的时候执行
            logger.info(
                f'start run output, plugin is {self._plugin_manager.output_plugin!r}.'
            )
            logger.complete()
            self.cluster.output(file_path, params)
        else:
            logger.info('未指定`output_plugin`, 跳过 `run_output` 函数.')

    def run(
        self,
        input_path: str = '',
        output_file_path: str = '',
        output_plugin_params: Dict[str, str] = {},
        path: str = '',
    ) -> Cluster:
        """执行所有插件

        Args:
            input_path: 输入路径，可以为空
            output_file_path: 输出文件路径, 可以为空
            output_plugin_params: 传递给output_plugin的参数, 可以为空
            path: 废弃参数

        Returns:
            Cluster: 集群对象
        """

        if path:
            import rich

            rich.print('[red]NetInspect.run: `path`参数已废弃, 请使用`input_path`参数代替.[/]')
            input_path = path

        self.run_input(input_path)
        self.run_parse()
        self.run_analysis()
        self.run_output(output_file_path, output_plugin_params)

        return self.cluster

    def add_device_with_raw_data(
        self, hostname: str, ip: str = '', cmd_contents: Dict[str, str] = {}
    ):
        """添加设备到集群中

        Args:
            hostname: 主机名
            ip: 管理ip, 可以为空
            cmd_contents: 命令内容
        """
        self.cluster.add_device_with_raw_data(hostname, ip, cmd_contents)

    def add_device(self, input_plugin_result: InputPluginResult):
        """添加设备到集群中

        Args:
            input_plugin_result: InputPluginResult的实例
        """
        self.cluster.add_device_use_input_plugin_result(input_plugin_result)

    def search(self, device_name: str) -> List[Device]:
        """搜索设备

        Args:
            device_name: 设备名称

        Returns:
            设备列表
        """

        return self.cluster.search(device_name)

    def get_base_info(self) -> List[BaseInfo]:
        """获取所有设备的基本信息

        Returns:
            List[BaseInfo]: 基本信息列表
        """

        ret = []
        for device in self.cluster.devices:
            ret.append(device.info)

        return ret

    def set_base_info_handler(self, handler: Type[EachVendorDeviceInfo]):
        """设置设备基本信息处理器

        Args:
            handler: 设备基本信息处理器
        """
        self.cluster.base_info_handler = handler()



from rich import print
from ..domain import OutputPluginAbstract


class OutputPluginWithDeviceWarningLogging(OutputPluginAbstract):
    """
    以漂亮的格式输出所有设备的告警信息
    """

    def main(self):
        for device in self.args.devices:
            for warn in device.analysis_result:
                if warn.above_focus:
                    print(
                        f"[green]%s[/] -- [{'red' if warn.is_warning else 'yellow'}]%s[/] -- %s"
                        % (device.info.hostname, warn.level_str, warn.message)
                    )



import os
from . import __file__ as module_file


class PyStr:
    def __init__(self):
        self.software = 'net_inspect'
        self.default_parse_plugin = 'ntc_templates'
        self.parse_plugin_prefix = 'parse --'
        self.analysis_plugin_prefix = 'analysis --'


class PyOption:
    def __init__(self):
        self.input_file_expend = ['.txt', '.log', '.diag']

        self.console_log_level = 'INFO'
        self.console_format = "{time:HH:mm:ss} | <level>{level}</level> | {message}"

        self.logfile_format = (
            "{time:YYYY-MM-DD HH:mm:ss} [<level>{level}</level>] | {name}: {message}"
        )
        self.logfile_name = 'net_inspect.log'
        self.logfile_rotation = '2 MB'


class PyPath:
    def __init__(self):
        self.root_path = os.path.dirname(module_file)
        self.project_path = os.path.dirname(self.root_path)


pystr = PyStr()
pyoption = PyOption()
pypath = PyPath()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['net_inspect', 'net_inspect.plugins', 'net_inspect.third_party']

package_data = \
{'': ['*']}

install_requires = \
['loguru>=0.6.0,<0.7.0',
 'ntc_templates_elinpf>=3.3.0,<4.0.0',
 'rich>=12.4.1,<13.0.0',
 'rich_typer>=0.1.5,<0.2.0',
 'textfsm>=1.1.2,<2.0.0']

entry_points = \
{'console_scripts': ['net_inspect = net_inspect.cli:app']}

setup_kwargs = {
    'name': 'net-inspect',
    'version': '1.5.2',
    'description': '基于已收集的网络设备信息进行的结构化数据分析框架',
    'long_description': '[![PyPI](https://img.shields.io/pypi/v/net_inspect?color=orange)](https://pypi.org/project/net-inspect)\n[![Python version](https://img.shields.io/badge/python-3.7%2B-blue)](https://github.com/Elinpf/net_inspect)\n[![GitHub stars](https://img.shields.io/github/stars/Elinpf/net_inspect)](https://github.com/Elinpf/net_inspect/stargazers)\n[![GitHub license](https://img.shields.io/github/license/Elinpf/net_inspect)](https://github.com/Elinpf/net_inspect/blob/master/LICENSE)\n[![codecov](https://codecov.io/gh/Elinpf/net_inspect/branch/master/graph/badge.svg?token=SVPWMH4XHY)](https://codecov.io/gh/Elinpf/net_inspect)\n[![Documentation Status](https://readthedocs.org/projects/net-inspect/badge/?version=latest)](https://net-inspect.readthedocs.io/zh/latest/?badge=latest)\n\n[net_inspect 教程](https://net-inspect.readthedocs.io/zh/latest/)\n\n## 前言\n\n作为网络工程师，需要经常收集设备的设备信息，比如设备CPU利用率、内存利用率、软件版本、序列号、接口状态等，这些信息可以用来监控网络设备的运行状态，也可以用来做设备的资产管理。\n\n在已经收集完成的情况下，我们需要将关键信息提取出来，然后将可用数据按需求格式输出。市面上目前没有针对多厂商的网络设备数据解析工具，厂家只会制作针对自己设备的解析工具。造成使用割裂，并且输出样式也无法做到自定义。\n\n另外，网络工程师学习编程后大部分都是单兵作战，针对自己的使用场景，进行专门的脚本编写，但这就导致了脚本的可复用性差、性能优化不足、脚本设计考虑不够周全、扩展性不强、调试困难等。\n\n通常情况下, 在获得设备回显信息后, 通过`textFSM`或者正则表达式解析数据，最后将可用的数据按需求格式进行输出。\n\n但是，这样的操作方式有很多不便，比如：\n\n1. 在通过`show`或者`display`将信息收集后，如何分割命令的回显？哪些是无效命令？\n2. 网管平台收集的信息格式不一样怎么处理？\n3. 如何识别设备对应的厂商？\n4. 如何调用对应的`textFSM`模板？\n5. 如何搜索命令获取关键信息？\n6. 多厂商的情况下，如何做到统一的数据调用？\n7. 多厂商的情况下，如何做到统一的数据格式？\n8. `ntc-templates`大部分模板都是国外厂商，怎么办？\n9. 如何在调用`ntc-templates`的同时调用自己写的`textFSM`模板库？\n10. 设计的脚本只针对单一情况，如何解决复用性的问题？\n\n要解决以上问题，就需要用到`net_inspect`这个框架。\n\n它可以让使用者无需关心数据处理的各种细节以及各个厂家的差异。屏蔽了厂商差异，只需要在数据处理完成后通过统一的方式调用即可。\n\n## net_inspect框架介绍\n\n[net_inspect](https://github.com/Elinpf/net_inspect)是一个基于`python`的网络设备数据解析自动化框架\n\n框架设计的初衷就是将各个环节解耦，使各个环节能够独立开发扩展，最终实现对多厂商设备的数据解析。\n\n安装方法：\n\n```bash\npip install net_inspect\n```\n\n### 一个简单的例子\n\n```\n├── log\n│\xa0\xa0 ├── Switch_A.log\n│\xa0\xa0 ├── Switch_B.log\n│\xa0\xa0 ├── Switch_C.log\n```\n\n可用看到，在`log`文件夹下有三台设备的命令的回显信息，这些文件是通过`vty`或者`console`连接到设备上作为记录日志保存下来。每个文件中都需要包含`display version` 这个命令来识别设备的厂商。\n\n现在我们需要将这些设备的厂商、软件版本、管理IP，提取打印出来。\n\n```py\nfrom net_inspect import NetInspect\n\nnet = NetInspect()\nnet.set_plugins(input_plugin=\'console\')\nnet.run(input_path=\'log\')\n\nprint(\'total devices:\', len(net.cluster.devices))\n\nfor device in net.cluster.devices:\n    info = device.info\n    print(\' | \'.join([info.hostname, info.ip, info.vendor,\n          info.version, info.model, info.cpu_usage]))\n```\n\noutput:\n\n```text\ntotal devices: 3\nSwitch_A | 24.44.1.248 | Huawei | 5.170 | S12704 Terabit Routing Switch | 6%\nSwitch_B | 24.45.1.240 | H3C | 5.20 Release: 3342 | SR8800 | 5%\nSwitch_C | 24.45.254.254 | H3C | 5.20 Release: 1238P08 | S9508E-V | 1%\n```\n\n可以看到，仅仅通过简单的几行代码，就可以实现对多厂商设备的数据解析和调用。\n\n> `console`代表的是`vty`或者`console`连接设备的回显信息格式。\n\n> 可供使用的base_info属性信息在使用`net_inspect -b`查看\n> 其中`analysis`的分析结果均是`Optional[bool]` 类型\n> 1. 如果为`None`，则表示未分析出结果。\n> 2. 如果为`True`，则表示有告警级别的信息。\n> 3. 如果为`False`，则表示为正常或者只是关注级别的信息。\n\n### 框架的组成\n\n`net_inspect`由以下几个部分组成：\n\n- `input_plugin` 输入插件，用于将设备的回显进行分割，将每个执行的命令和对应的回显作为字典储存。\n- `parse_plugin` 解析插件，用于对命令进行解析，一般情况下不需要进行修改。\n- `analysis_plugin` 分析插件，每一个分析插件对应一个状态分析，比如`power_status`就是用来分析设备的电源状态的。\n- `output_plugin` 输出插件，用于输出统一的格式报告。\n- `base_info` 基础信息，用于存放设备的基础信息，比如厂商、软件版本、管理IP等。可以全厂商统一调用。\n\n> 有时候我们收集设备信息的方式不止是通过console，或者自己的telnet/ssh。而是通过网管平台进行的，回显格式不同，这时候就需要自己写一个`input_plugin`，将网管平台的数据转换成`net_inspect`需要的格式。\n\n> `parse_plugin` 使用的是[ntc_templates_elinpf](https://github.com/Elinpf/ntc-templates)这个库，它是`ntc_templates`的一个分支，主要是为了解决`ntc_templates`没有对国内厂商支持的问题。\n\n> 所以，需要先卸载`pip uninstall ntc_templates` 然后安装`pip install ntc_templates_elinpf`\n\n`net_inspect`支持cli的命令行模式，可以通过`net_inspect -h`查看帮助信息。\n\n![](asset/2.png)\n\n\n## 进阶教程\n\n上面的简单例子中还有很多地方没有涉及到，比如`analysis_plugin`的使用，`base_info`的使用，`output_plugin`的使用等等。\n\n试想一个情况，我们需要获取设备的`show clock`信息，在`ntc_templates_elinpf`中没有，怎么办？\n\n### 调用外部textFSM模板\n\n要解决`ntc_templates_elinpf`中没有模板的问题，我们可以自己写一个本地的textFSM模板仓库，然后调用。\n\n```text\n├── local_templates\n│\xa0\xa0 ├── huawei_vrp_display_clock.textfsm\n│\xa0\xa0 ├── hp_comware_display_clock.textfsm\n│\xa0\xa0 ├── cisco_ios_show_clock.textfsm\n│\xa0\xa0 ├── index\n```\n\n可以看到上面的本地模板库中，包含了`huawei_vrp`、`hp_comware`、`cisco_ios`三个厂商的`display clock`模板。\n\n另外还有一个`index`文件，用于指定模板的调用命令，毕竟我们在敲命令的时候不会敲全，比如`display clock`，我们可以敲`dis clo`，这时候就需要`index`文件来指定。\n\n`index`文件内容如下:\n\n```text\nTemplate, Hostname, Platform, Command\n\nhuawei_vrp_display_clock.textfsm, .*, huawei_vrp, dis[[play]] clo[[ck]]\n\nhp_comware_display_clock.textfsm, .*, hp_comware, di[[splay]] clo[[ck]]\n\ncisco_ios_show_clock.textfsm, .*, cisco_ios, sh[[ow]] clo[[ck]]\n```\n\n注意每个厂商需要空格隔开, 并且开头必须要有`Template, Hostname, Platform, Command`这行。\n\n做好准备工作后，只需要在脚本中调用本地模板库即可。\n\n```python\nfrom net_inspect import NetInspect, vendor\n\nnet = NetInspect()\nnet.set_plugins(input_plugin=\'console\')\nnet.set_external_templates(\'local_templates\')  # 调用本地模板库\nnet.run(input_path=\'log\')\n\nprint(\'total devices:\', len(net.cluster.devices))\n\nfor device in net.cluster.devices:\n    info = device.info\n    clock = \'\'\n    if device.vendor == vendor.Huawei:\n        with device.search_cmd(\'display clock\') as cmd:\n            if cmd.parse_result:\n                ps = cmd.parse_result[0]\n                clock = f"{ps[\'year\']}-{ps[\'month\']}-{ps[\'day\']} {ps[\'time\']}"\n\n    if device.vendor == vendor.H3C:\n        with device.search_cmd(\'display clock\') as cmd:\n            if cmd.parse_result:\n                ps = cmd.parse_result[0]\n                clock = f"{ps[\'year\']}-{ps[\'month\']}-{ps[\'day\']} {ps[\'time\']}"\n\n    print(\' | \'.join([info.hostname, clock]))\n```\n\noutput:\n\n```text\ntotal devices: 3\nSwitch_A | 2021-03-19 10:23:08\nSwitch_B | 2021-03-19 10:24:17\nSwitch_C | 2021-03-19 10:32:17\n```\n\n可以看到，我们通过调用本地模板库，成功解析了`display clock`命令。\n\n然后搜索厂商对应的命令，将解析的结果格式化为我们想要的格式。最后再输出。\n\n这样的写法是否有点麻烦，并且不利于复用呢？有没有更好的方式呢？\n\n### 新增基础信息\n\n我们可以在`base_info`中新增一个`clock`字段，然后只需要调用这个`clock`属性就可以了，方法如下：\n\n```python\nfrom net_inspect import NetInspect, EachVendorDeviceInfo, BaseInfo, Device\n\n\nclass AppendClock(BaseInfo):\n    clock: str = \'\'  # 巡检时间\n\n\nclass EachVendorWithClock(EachVendorDeviceInfo):\n\n    base_info_class = AppendClock  # 基本信息类\n\n    def do_huawei_vrp_baseinfo_2(self, device: Device, info: AppendClock):\n        # 添加do_<vendor_platform>_baseinfo_<something>方法，可以自动运行\n        with device.search_cmd(\'display clock\') as cmd:\n            if cmd.parse_result:\n                row = cmd.parse_result[0]\n                info.clock = f\'{row["year"]}-{row["month"]}-{row["day"]} {row["time"]}\'\n\n    def do_hp_comware_baseinfo_2(self, device: Device, info: AppendClock):\n        with device.search_cmd(\'display clock\') as cmd:\n            if cmd.parse_result:\n                row = cmd.parse_result[0]\n                info.clock = f\'{row["year"]}-{row["month"]}-{row["day"]} {row["time"]}\'\n\n\nif __name__ == \'__main__\':\n    net = NetInspect()\n    net.set_plugins(input_plugin=\'console\')\n    net.set_base_info_handler(EachVendorWithClock)  # 设置获取设备基本信息的处理类\n    net.set_external_templates(\'local_templates\')\n    net.run(input_path=\'log\')\n\n    print(\'total devices:\', len(net.cluster.devices))\n\n    for device in net.cluster.devices:\n        info = device.info  # type: AppendClock\n        print(\' | \'.join([info.hostname, info.clock]))\n```\n\noutput:\n\n```text\ntotal devices: 3\nSwitch_A | 2021-03-19 10:23:08\nSwitch_B | 2021-03-19 10:24:17\nSwitch_C | 2021-03-19 10:32:17\n```\n\n输出的结果是一样的，可以看到，在`base_info`中新增了一个`clock`字段，然后只需要调用这个`clock`属性就可以了。这样的做法复用性强，后续再想要获取设备的时间的时候，只需要调用`clock`属性即可。\n\n### 自定义分析信息\n\n完成了数据的解析调用，下面还需要对设备进行分析，比如判断电源、风扇是否故障，cpu、memory利用率是否过高等。\n\n当`net_inspect`中没有需要的分析模块时，我们可以自定义分析模块。\n\n比如我们写一个检查OSPF邻居状态的分析模块\n\n```py\nfrom __future__ import annotations\n\nfrom typing import TYPE_CHECKING\nfrom net_inspect import NetInspect, vendor\nfrom net_inspect.analysis_plugin import analysis, AnalysisPluginAbc\n\nif TYPE_CHECKING:\n    from net_inspect.analysis_plugin import TemplateInfo\n    from net_inspect.domain import AnalysisResult\n\n\nclass AnalysisPluginWithOSPFStatus(AnalysisPluginAbc):\n    """OSPF status 状态不能为Init"""\n\n    @analysis.vendor(vendor.Huawei)\n    @analysis.template_key(\'huawei_vrp_display_ospf_peer_brief.textfsm\', [\'neighbor\', \'state\'])\n    def huawei_vrp(template: TemplateInfo, result: AnalysisResult):\n        """华为状态检查"""\n        for row in template[\'display ospf peer brief\']:\n            if row[\'state\'].lower() == \'init\':\n                result.add_warning(f\'{row["neighbor"]} is in init state\')\n\n\nif __name__ == \'__main__\':\n    net = NetInspect()\n    net.set_plugins(input_plugin=\'console\')\n    net.run(input_path=\'log\')\n\n    print(\'total devices:\', len(net.cluster.devices))\n\n    for device in net.cluster.devices:\n        ospf_status = device.analysis_result.get(\'ospf status\')\n        warning_list = []\n        for alarm in ospf_status:\n            if alarm.is_warning:\n                warning_list.append(alarm.message)\n\n        print(\' | \'.join([device.info.hostname, \', \'.join(warning_list)]))\n```\n\noutput:\n\n```text\ntotal devices: 3\nSwitch_A | 24.42.254.20 is in init state\nSwitch_B |\nSwitch_C |\n```\n\n这样我们就得到了一个可重复利用的分析模块，后续在想要检查OSPF邻居状态的时候，只需要调用这个模块即可。\n\n这里的案例只添加了`Huawei`这一个厂家的分析方法，其他厂家的分析方法大家可以自行添加。\n\n编写这个需要注意：\n1. `@analysis.template_key`中的`templates`名称需要完整包含后缀名。\n2. 类方法不需要`self`这个关键字。\n3. 结果加入到`result`中即可，不需要返回值。\n4. 整个过程不用单独设置，所有信息会直接写入`analysis`全局变量中。\n5. 类注释和方法注释必须要写，因为会用到这个注释作为分析结果的标题。\n\n\n### 自定义输出模块\n\n完成了数据的分析后，需要对结果进行调用，可以直接以写脚本的方式，也可以自定义输出模块。\n\n自定义输出模块的方式方便二次调用，比如我们写一个输出table的例子：\n\n```py\nfrom net_inspect import NetInspect, OutputPluginAbstract\nfrom rich.table import Table\nfrom rich.console import Console\n\n\nclass Output(OutputPluginAbstract):\n    def main(self):\n        self.check_args(\'title\')  # 检查是否提供title参数\n        console = Console()\n\n        table = Table(title=self.args.output_params.get(\n            \'title\'), show_lines=False)\n        columns = [\'hostname\', \'ip\', \'model\', \'version\', \'power status\']\n        for col in columns:\n            table.add_column(col, justify=\'center\')\n        table.row_styles = [\'green\']\n\n        for device in self.args.devices:\n            info = device.info\n            table.add_row(\n                info.hostname,\n                info.ip,\n                info.model,\n                info.version,\n                \'Abnormal\' if info.analysis.power else \'Normal\'\n            )\n\n        console.print(table)\n\nif __name__ == \'__main__\':\n\n    net = NetInspect()\n    net.set_plugins(input_plugin=\'console\', output_plugin=Output)\n    cluster = net.run(input_path=\'log\', output_plugin_params={\n                    \'title\': \'设备信息\'})\n```\n\noutput:\n\n![](asset/1.png)\n\n可以看到，我们自定义了一个输出模块，可以直接调用，而不需要写脚本。\n\n`output_plugin_params`中的所有参数都会传递到`Output`类中的`output_params`变量中，可以直接使用。\n\n### 手动添加设备\n\n有时候我们需要手动添加设备，比如我们需要添加一个设备，但是这个设备没有日志，我们可以手动添加，并且指定设备的厂家，这样就可以使用对应厂家的分析模块。\n\n```py\nfrom net_inspect import NetInspect, InputPluginResult, vendor\n\nnet = NetInspect()\n\nd = InputPluginResult()\nd.add_cmd(\'display clock\', "2021-03-19 10:23:08+08:00")\nd.hostname = \'Device\'\nd.vendor = vendor.Huawei\n\nnet.add_device(d)\n\nnet.run()\n\nfor device in net.cluster.devices:\n    print(device.info.hostname)\n    print(device.parse_result(\'dis clo\'))\n```\n\noutput:\n\n```text\nDevice\n[{\'time\': \'10:23:08\', \'timezone\': \'\', \'dayweek\': \'\', \'year\': \'2021\', \'month\': \'03\', \'day\': \'19\'}]\n```\n\n## 关于贡献\n\n分析插件还在持续开发中，`develop_script.py`脚本就是为高效开发提供的一个工具。\n\n开发一个分析插件的流程，以开发检查风扇状态的`fan_status`插件为例：\n\n1. 创建一个新的插件文件, 对应的文件初始状态会一并准备好\n\n```bash\npython ./develop_script.py -p fan_status -g\n```\n\n2. 在对应的文件中实现插件对每个厂商分析的函数\n\n```py\nclass AnalysisPluginWithFanStatus(AnalysisPluginAbc):\n    """\n    要求设备所有在位风扇模块运行在正常状态。\n    """\n    @analysis.vendor(vendor.H3C)\n    @analysis.template_key(\'hp_comware_display_fan.textfsm\', [\'slot\', \'id\', \'status\'])\n    def hp_comware(template: TemplateInfo, result: AnalysisResult):\n        """模块状态不为Normal的时候告警"""\n        for row in template[\'display fan\']:\n            if row[\'status\'].lower() != \'normal\':\n                result.add_warning(\n                    f\'Slot {row["slot"]} Fan {row["id"]} 状态异常\' if row[\'slot\'] else f\'Fan {row["id"]} 状态异常\')\n```\n\n其中`@analysis`是用来记录插件的分析类型的，`vendor`记录插件的厂商类型，`template_key`记录分析模块所需要的`textfsm`文件以及里面的哪些值。\n这些值会在参数`template: TemplateInfo`中给出。\n\n`result: AnalysisResult`用来记录分析结果。可以添加告警信息。\n\n分析方法为类方法，不需要`self`,不需要给出返回值。\n\n插件中的类注释和方法注释都会被记录下来，方便后续调用。\n\n3. 创建对应的测试文件\n\n当编写了对应的分析方法后，再次执行创建命令，工具会自动根据分析方法中需要的命令，生成对应的测试文件。\n\n测试文件路径为`tests/check_analysis_plugins/<plugin_name>/<funcation_name>.raw`\n\n```bash\npython ./development_script.py -p fan_status -f hp_comware -g\n```\n\n4. 在测试文件中添加测试用例\n5. 执行测试\n\n```bash\npython ./development_script.py -p fan_status -f hp_comware -t\n```\n\n6. 完成测试，确认测试结果为正常后，生成yml文件作为参考文件。\n\n```bash\npython ./development_script.py -p fan_status -f hp_comware -y\n```\n\n',
    'author': 'Elin',
    'author_email': '365433079@qq.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Elinpf/net_inspect',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

from .. import vendor
from ..analysis_plugin import AnalysisPluginAbc, analysis
from ..func import safe_str2float

if TYPE_CHECKING:
    from ..analysis_plugin import TemplateInfo
    from ..domain import AnalysisResult


class AnalysisPluginWithMemoryStatus(AnalysisPluginAbc):
    """
    检查内存利用率不能过高。
    """

    @analysis.vendor(vendor.Huawei)
    @analysis.base_info('memory_usage')
    def huawei_vrp(template: TemplateInfo, result: AnalysisResult):
        """
        内存利用率高于80%，则告警。
        """
        check_memory_usage(template.base_info['memory_usage'], result)

    @analysis.vendor(vendor.Cisco)
    @analysis.base_info('memory_usage')
    def cisco_ios(template: TemplateInfo, result: AnalysisResult):
        """
        内存利用率高于80%，则告警。
        """
        check_memory_usage(template.base_info['memory_usage'], result)

    @analysis.vendor(vendor.Ruijie)
    @analysis.base_info('memory_usage')
    def ruijie_os(template: TemplateInfo, result: AnalysisResult):
        """
        内存利用率高于80%，则告警。
        """
        check_memory_usage(template.base_info['memory_usage'], result)

    @analysis.vendor(vendor.Maipu)
    @analysis.base_info('memory_usage')
    def maipu_mypower(template: TemplateInfo, result: AnalysisResult):
        """
        内存利用率高于80%，则告警。
        """
        check_memory_usage(template.base_info['memory_usage'], result)

    @analysis.vendor(vendor.H3C)
    @analysis.base_info('memory_usage')
    def hp_comware(template: TemplateInfo, result: AnalysisResult):
        """
        内存利用率高于80%，则告警。
        """
        check_memory_usage(template.base_info['memory_usage'], result)


def check_memory_usage(memory_usage: str, result: AnalysisResult):
    if memory_usage == '':
        return

    # 去除百分号，转换为float
    memory_usage = safe_str2float(memory_usage[:-1])

    if memory_usage > 80:
        result.add_warning(f"内存利用率达到{round(memory_usage,1)}%")



from __future__ import annotations

from typing import TYPE_CHECKING

from .. import vendor
from ..analysis_plugin import AnalysisPluginAbc, analysis
from ..func import match_lower

if TYPE_CHECKING:
    from ..analysis_plugin import TemplateInfo
    from ..domain import AnalysisResult


class AnalysisPluginWithPowerStatus(AnalysisPluginAbc):
    """
    要求设备所有在位电源模块运行在正常状态。
    """

    @analysis.vendor(vendor.Huawei)
    @analysis.template_key(
        'huawei_vrp_display_power.textfsm', ['mode', 'id', 'present', 'state']
    )
    def huawei_vrp(template: TemplateInfo, result: AnalysisResult):
        """模块状态在线，并且状态不为Normal或者Supply时告警"""
        power = template['display power']
        for row in power:
            # 当没有电源插入的时候，不关注
            if not match_lower(row['present'], r'yes|present'):
                result.add_focus(f"Power {row['id']} 电源不在位")
            elif not match_lower(row['state'], r'normal|supply'):
                result.add_warning(f'Power {row["id"]} 电源状态异常')

    @analysis.vendor(vendor.Cisco)
    @analysis.template_key(
        'cisco_ios_show_power_status.textfsm', ['ps', 'model', 'status', 'fan_sensor']
    )
    def cisco_ios(template: TemplateInfo, result: AnalysisResult):
        """模块状态不为good时，报警"""
        for row in template['show power status']:
            if row['status'] != 'good':
                result.add_warning(f'{row["ps"]}: {row["model"]} 电源状态异常')

    @analysis.vendor(vendor.Ruijie)
    @analysis.template_key(
        'ruijie_os_show_power.textfsm', ['power_id', 'power_type', 'status']
    )
    def ruijie_os(template: TemplateInfo, result: AnalysisResult):
        """模块状态不为ok时，报警"""
        for row in template['show power']:
            if row['power_type'] == 'N/A':
                continue
            if row['status'] != 'ok':
                result.add_warning(f'Power {row["power_id"]} 电源状态异常')

    @analysis.vendor(vendor.Maipu)
    @analysis.template_key(
        'maipu_mypower_show_system_power.textfsm',
        ['power_id', 'status', 'work_status', 'power_in'],
    )
    def maipu_mypower(template: TemplateInfo, result: AnalysisResult):
        """
        迈普的show system power 分为两种:
        一种为长的信息，``status``的信息为online， ``work_status`` 和 ``power_in`` 均为Normal为正常，
        另外一种是短信息，只有 ``status``, 判断模块是否正常
        """
        for row in template['show system power']:
            if row['status'].lower() == 'online':
                if (
                    row['work_status'].lower() != 'normal'
                    or row['power_in'].lower() != 'normal'
                ):
                    result.add_warning(f'Power {row["power_id"]} 状态异常')
            else:
                if row['status'].lower() != 'normal':
                    result.add_warning(f'Power {row["power_id"]} 状态异常')

    @analysis.vendor(vendor.H3C)
    @analysis.template_key('hp_comware_display_power.textfsm', ['slot', 'id', 'status'])
    def hp_comware(template: TemplateInfo, result: AnalysisResult):
        """模块状态不为Normal的时候告警"""
        for row in template['display power']:
            if row['status'].lower() != 'normal':
                result.add_warning(
                    f'Slot {row["slot"]} Power {row["id"]} 状态异常'
                    if row['slot']
                    else f'{row["id"]} 状态异常'
                )
